
import os
import unittest
import unittest.mock as mock

import pytest

from primrose.configuration.configuration import Configuration
from primrose.readers.csv_reader import CsvReader
from primrose.dag_runner import DagRunner
from testfixtures import LogCapture
from primrose.node_factory import NodeFactory
from primrose.writers.abstract_file_writer import AbstractFileWriter
from primrose.dag.config_layer_traverser import ConfigLayerTraverser
from primrose.data_object import DataObject
from primrose.data_object import DataObjectResponseType
from primrose.dag.dag_traverser import DagTraverser
from primrose.dag.traverser_factory import TraverserFactory
from abc import abstractmethod
from primrose.base.writer import  AbstractWriter


def test_run():
    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)
    runner.run()

def test_run2():
    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)
    with LogCapture() as l:
        runner.run(dry_run=True)
    l.check(('root', 'INFO', 'Taking nodes to run from default'),
            ('root', 'INFO', 'DRY RUN 0: would run node csv_reader of type reader_config and class CsvReader'),
            ('root', 'INFO', 'All done. Bye bye!'))

def test_run3():
    config = {
        "metadata": "ConfigLayerTraverser",
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["csv_writer"]
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "test/unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)

    with LogCapture() as l:
        runner.run(dry_run=True)
    l.check(('root', 'INFO', 'Taking nodes to run from default'),
            ('root', 'INFO', 'DRY RUN 0: would run node csv_reader of type reader_config and class CsvReader'),
            ('root', 'INFO', 'DRY RUN 1: would run node csv_writer of type writer_config and class CsvWriter'),
            ('root', 'INFO', 'All done. Bye bye!'))

def test_run4():

    class TestWriter(AbstractFileWriter):
        def __init__(self, configuration, instance_name):
            pass
        @staticmethod
        def necessary_config(node_config):
            return set([])
        def run(self, data_object):
            terminate = True
            return data_object, terminate
    NodeFactory().register("TestWriter", TestWriter)

    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["csv_writer"]
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "TestWriter"
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)

    with LogCapture() as l:
        runner.run(dry_run=False)
    l.check(('root', 'INFO', 'Taking nodes to run from default'),
        ('root','INFO','received node csv_reader of type reader_config and class CsvReader'),
        ('root', 'INFO', 'Reading test/minimal.csv from CSV'),
        ('root', 'INFO', 'received node csv_writer of type writer_config and class TestWriter'),
        ('root', 'INFO', 'Terminating early due to signal from csv_writer'),
        ('root', 'INFO', 'All done. Bye bye!'))

def test_run5():
    config = {
        "metadata": {
            "section_registry":[
                "phase1",
                "phase2"
            ]
        },
        "implementation_config":{
            "phase1": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["csv_writer"]
                }
            },
            "phase2": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "test/unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)

    with LogCapture() as l:
        runner.run(dry_run=True)
    l.check(('root', 'INFO', 'Taking nodes to run from section_registry'),
            ('root', 'INFO', 'DRY RUN 0: would run node csv_reader of type phase1 and class CsvReader'),
            ('root', 'INFO', 'DRY RUN 1: would run node csv_writer of type phase2 and class CsvWriter'),
            ('root', 'INFO', 'All done. Bye bye!'))


def test_run6():
    config = {
        "metadata": {
            "section_registry": [
                "phase1",
                "cleanup_config"
            ],
            "notify_on_error": {
                "client": "SlackClient",
                "channel": "some-channel",
                "token": "slack-api-token",
                "member_id": "optional-key"
            }
        },
        "implementation_config": {
            "phase1": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["notification"]
                }
            },
            "cleanup_config": {
                "notification": {
                    "class": "ClientNotification",
                    "client": "SlackClient",
                    "channel": "some-channel",
                    "token": "slack-api-token",
                    "member_id": "optional-key",
                    "message": "Yay! Sucess"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)

    with LogCapture() as l:
        runner.run(dry_run=True)
    l.check(('root', 'INFO', 'Taking nodes to run from section_registry'),
            ('root', 'INFO', 'DRY RUN 0: would run node csv_reader of type phase1 and class CsvReader'),
            ('root', 'INFO', 'DRY RUN 1: would run node notification of type cleanup_config and class ClientNotification'),
            ('root', 'INFO', 'All done. Bye bye!'))


def test_run_notification_error():
    config = {
        "metadata": {
            "section_registry": [
                "phase1"
            ],
            "notify_on_error": {
                "client": "SlackClient",
                "channel": "some-channel",
                "token": "slack-api-token",
                "member_id": "optional-key"
            }
        },
        "implementation_config": {
            "phase1": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "bad/path.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)

    mock_client = mock.Mock()
    mock_client.post_message = mock.Mock()
    mock_get_notification_client = mock.Mock(return_value=mock_client)

    path = 'primrose.notification_utils.get_notification_client'
    with mock.patch(path) as mock_get_notification_client:
        with pytest.raises(Exception) as error:
            runner.run()
            assert mock_get_notification_client.post_message.call_count == 1


def test_cache_data_object():
    config = {
        "metadata": {
            "data_object": {
                "write_to_cache": True,
                "write_filename": "dag_runner_test_cache_data_object.pkl"
            }
        },
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)
    writer = CsvReader(configuration, "csv_reader")
    data_object.add(writer, "some_data")

    runner = DagRunner(configuration)

    filename = "dag_runner_test_cache_data_object.pkl"
    if os.path.exists(filename):
        os.remove(filename)

    cached = runner.cache_data_object(data_object)

    assert cached

    assert os.path.exists(filename)

    if os.path.exists(filename):
        os.remove(filename)

def test_create_data_object():

    filename = "dag_runner_create_data_object.pkl"
    # hack part 1: make sure this filename exists so that checks in Configuration pass
    open(filename, 'w+')

    config = {
        "metadata": {
            "data_object": {
                "read_from_cache": True,
                "read_filename": "dag_runner_create_data_object.pkl"
            }
        },
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    # hack part 2: now get rid of it
    if os.path.exists(filename):
        os.remove(filename)

    # now write the actual object to restore from
    data_object = DataObject(configuration)
    writer = CsvReader(configuration, "csv_reader")
    data_object.add(writer, "some_data")
    data_object.write_to_cache(filename)
    assert os.path.exists(filename)

    # now we get to the code to test
    runner = DagRunner(configuration)
    restored_data_object = runner.create_data_object()

    # run some checks
    assert isinstance(restored_data_object, DataObject)
    assert restored_data_object.get("csv_reader", rtype=DataObjectResponseType.VALUE.value) == "some_data"

    #cleanup
    if os.path.exists(filename):
        os.remove(filename)

def test_init_traverser_from_config():

    class TestTraverser(DagTraverser):
        def traversal_list(self): return []
        def run_section_by_section(self): return False

    TraverserFactory().register("TestTraverser", TestTraverser)

    config = {
        "metadata": {
            "traverser": "TestTraverser"
        },
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    assert isinstance(runner.dag_traverser, TestTraverser)

def test_filter_sequence1():
    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    with pytest.raises(Exception) as e:
        runner.filter_sequence(["csv_reader", "csv_reader"])
    assert "You have duplicate nodes from traverser!" in str(e)

def test_filter_sequence2():
    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    with pytest.raises(Exception) as e:
        runner.filter_sequence(["csv_reader", "junk!"])
    assert "Unknown key junk!" in str(e)

def test_filter_sequence3():
    config = {
        "implementation_config":{
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    with pytest.raises(Exception) as e:
        runner.filter_sequence([])
    assert "Ran out of nodes for section reader_config. Only received []" in str(e)

def test_filter_sequence4():
    config = {
        "metadata": {
            "section_run": [
                "writer_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["recipe_csv_writer"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    sequence = runner.filter_sequence(["read_data","recipe_csv_writer"])
    assert sequence == ["recipe_csv_writer"]

def test_filter_sequence5():
    config = {
        "metadata": {
            "section_run": [
                "writer_config",
                "reader_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["recipe_csv_writer","recipe_csv_writer2"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                },
                "recipe_csv_writer2": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)
    with pytest.raises(Exception) as e:
        runner.filter_sequence(["recipe_csv_writer", "read_data", "recipe_csv_writer2"])
    assert "Traverser is mismatched with section writer_config. Expecting set ['recipe_csv_writer', 'recipe_csv_writer2']" in str(e)

def test_filter_sequence6():
    #test of dependencies with a section
    config = {
        "metadata": {
            "traverser": "DepthFirstTraverser"
        },
        "implementation_config": {
            "reader_config": {
                "read_data1": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["csv_writer"]
                },
                "read_data2": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["csv_writer"]
                },
                "read_data3": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["read_data2"]
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)
    s_in = ["read_data3", "read_data2", "read_data1", "csv_writer"]
    s_out = runner.filter_sequence(s_in)
    assert s_out == s_in

    with pytest.raises(Exception) as e:
        runner.filter_sequence([ "read_data1", "csv_writer", "read_data3", "read_data2"])
    assert "Upstream path found, from read_data3 to csv_writer" in str(e.value)

def test_check_upstream():
    config = {
        "metadata": {
            "section_run": [
                "reader_config",
                "writer_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["recipe_csv_writer","recipe_csv_writer2"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                },
                "recipe_csv_writer2": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)
    assert runner.check_for_upstream(['read_data', 'recipe_csv_writer', 'recipe_csv_writer2']) == False

def test_check_upstream2():
    config = {
        "metadata": {
            "section_run": [
                "writer_config",
                "reader_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["recipe_csv_writer","recipe_csv_writer2"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                },
                "recipe_csv_writer2": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)
    with pytest.raises(Exception) as e:
        runner.check_for_upstream(['recipe_csv_writer2', 'read_data', 'recipe_csv_writer'])
    assert "Upstream path found, from read_data to recipe_csv_writer2" in str(e.value)


def test_run_bad():
    class TestWriterTmp(AbstractWriter):
        @staticmethod
        def necessary_config(node_config):
            return set([])

        def run(self, data_object):
            return data_object, False

    NodeFactory().register("TestWriterTmp", TestWriterTmp)

    config = {
        "implementation_config":{
            "writer_config": {
                "mywriter": {
                    "class": "TestWriterTmp",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    # unregister this class
    del NodeFactory().name_dict["TestWriterTmp"]

    with pytest.raises(Exception) as e:
        runner.run()
    assert "Issue instantiating mywriter and class TestWriterTmp" in str(e)

def test_run_bad2():
    class TestWriterTmp(AbstractWriter):
        @staticmethod
        def necessary_config(node_config):
            return set([])

        def run(self, data_object):
            raise Exception("Deliberate error")
            #return data_object, False

    NodeFactory().register("TestWriterTmp", TestWriterTmp)

    config = {
        "implementation_config":{
            "writer_config": {
                "mywriter": {
                    "class": "TestWriterTmp",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    runner = DagRunner(configuration)

    with pytest.raises(Exception) as e:
        runner.run()
    assert "Deliberate error" in str(e)

def test_run_pruned():
    config = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "data/tennis.csv",
                    "destinations": ["conditional_node"]
                },
                "conditional_node": {
                    "class": "SimpleSwitch",
                    "path_to_travel": "left",
                    "destinations": ['left', 'right']
                }
            },
            "writer_config": {
                "left": {
                    "class": "LoggingSuccess",
                    "msg": "left node!",
                    "level": "INFO",
                },
                "right": {
                    "class": "LoggingSuccess",
                    "msg": "right node!",
                    "level": "INFO",
                    "destinations": ["right2"]
                },
                "right2": {
                    "class": "LoggingSuccess",
                    "msg": "right node2!",
                    "level": "INFO",
                }
            }
        }
    }
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    runner = DagRunner(configuration)
    with LogCapture() as l:
        runner.run()
    l.check(('root', 'INFO', 'Taking nodes to run from default'),
            ('root', 'INFO', 'received node read_data of type reader_config and class CsvReader'),
            ('root', 'INFO', 'Reading data/tennis.csv from CSV'),
            ('root', 'INFO', 'received node conditional_node of type reader_config and class SimpleSwitch'),
            ('root', 'INFO', 'Skipping pruned node right'),
            ('root', 'INFO', 'Skipping pruned node right2'),
            ('root', 'INFO', 'received node left of type writer_config and class LoggingSuccess'),
            ('root', 'INFO', 'left node!'),
            ('root', 'INFO', 'All done. Bye bye!'))



import pytest

from primrose.pipelines.sklearn_preprocessing_pipeline import SklearnPreprocessingPipeline
from primrose.transformers.sklearn_preprocessing_transformer import SklearnPreprocessingTransformer

def test__instantiate_preprocessor():
    processor = SklearnPreprocessingPipeline._instantiate_preprocessor("preprocessing.StandardScaler", args=None)
    assert isinstance(processor, SklearnPreprocessingTransformer)

    with pytest.raises(Exception) as e:
        SklearnPreprocessingPipeline._instantiate_preprocessor("preprocessing.junk", args=None)
    assert "Preprocessor junk not found in preprocessing module" in str(e)




import dill
import pickle
import pytest
import os
from sklearn.tree import DecisionTreeClassifier

from primrose.configuration.configuration import Configuration
from primrose.data_object import DataObject, DataObjectResponseType
from primrose.readers.csv_reader import CsvReader
from primrose.readers.deserializer import Deserializer, GcsDeserializer


def test_init_ok_dill():
    config = {
        "implementation_config": {
            "reader_config": {
                "dill_reader": {
                    "class": "Deserializer",
                    "filename": "test/tinymodel.dill",
                    "deserializer": 'dill',
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = Deserializer(configuration, "dill_reader")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    data = data_object.get('dill_reader', rtype=DataObjectResponseType.VALUE.value)

    assert data is not None
    assert set(data.keys()) == {'test', 'model'}

    node_config = {
                    "class": "Deserializer",
                    "filename": "test/tinymodel.dill",
                    "deserializer": 'dill',
                    "destinations": []
                }

    assert isinstance(Deserializer.necessary_config(node_config), set)
    assert len(Deserializer.necessary_config(node_config)) > 0

    assert data['test'] == [1, 2, 3]
    assert isinstance(data['model'], DecisionTreeClassifier)


def test_init_ok_pickle():
    config = {
        "implementation_config": {
            "reader_config": {
                "pickle_reader": {
                    "class": "Deserializer",
                    "filename": "test/tinymodel.pickle",
                    "deserializer": 'pickle',
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = Deserializer(configuration, "pickle_reader")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    data = data_object.get('pickle_reader', rtype=DataObjectResponseType.VALUE.value)

    assert data is not None
    assert set(data.keys()) == {'test', 'model'}

    assert data['test'] == [1, 2, 3]
    assert isinstance(data['model'], DecisionTreeClassifier)

def test_init_ok_unsupported():
    config = {
        "implementation_config": {
            "reader_config": {
                "other_reader": {
                    "class": "Deserializer",
                    "filename": "test/tinymodel.pickle",
                    "deserializer": 'other',
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = Deserializer(configuration, "other_reader")
    with pytest.raises(Exception, match=r"Unsupported"):
        reader.run(data_object)

def test_gcsdeserializer_necessary_config():
    assert len(GcsDeserializer.necessary_config({})) == 3

def test_run_dill(monkeypatch):
    # returns 2 objects from dill reader
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDeserializer",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "deserializer": "dill",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDeserializer(configuration, "myreader")

    data_object = DataObject(configuration)

    def fake_blobs():
        with open("test/dill_reader_blob1.pkl", "wb") as dill_file:
            dill.dump("some_data", dill_file)
        with open("test/dill_reader_blob2.pkl", "wb") as dill_file:
            dill.dump("some_other_data", dill_file)
        dat1 = open("test/dill_reader_blob1.pkl", "rb").read()
        dat2 = open("test/dill_reader_blob2.pkl", "rb").read()
        return [dat1, dat2]

    monkeypatch.setattr(reader,'download_blobs_as_strings',fake_blobs)

    reader_object, terminate = reader.run(data_object)

    assert not terminate

    assert "myreader" in reader_object.data_dict

    dat = reader_object.data_dict['myreader']

    assert "reader_data" in dat
    datlist = dat['reader_data']

    assert len(datlist) == 2

    assert "some_data" in datlist
    assert "some_other_data" in datlist

    files = ["test/dill_reader_blob1.pkl", "test/dill_reader_blob2.pkl"]
    for f in files:
        if os.path.exists(f):
            os.remove(f)

def test_run_dill_2(monkeypatch):
    # returns 1 objects from dill reader
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDillReader",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "deserializer": "dill",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDeserializer(configuration, "myreader")

    data_object = DataObject(configuration)

    def fake_blobs():
        with open("test/dill_reader_blob1.pkl", "wb") as dill_file:
            dill.dump("some_data", dill_file)
        dat1 = open("test/dill_reader_blob1.pkl", "rb").read()
        return [dat1]

    monkeypatch.setattr(reader,'download_blobs_as_strings',fake_blobs)

    reader_object, terminate = reader.run(data_object)

    assert not terminate

    assert "myreader" in reader_object.data_dict

    dat = reader_object.data_dict['myreader']

    assert "reader_data" in dat
    data = dat['reader_data']

    assert "some_data" == data

    files = ["test/dill_reader_blob1.pkl"]
    for f in files:
        if os.path.exists(f):
            os.remove(f)

def test_run_pickle(monkeypatch):
    # returns 2 objects from pickle reader
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDeserializer",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "deserializer": "pickle",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDeserializer(configuration, "myreader")

    data_object = DataObject(configuration)

    def fake_blobs():
        with open("test/pickle_reader_blob1.pkl", "wb") as pickle_file:
            pickle.dump("some_data", pickle_file)
        with open("test/pickle_reader_blob2.pkl", "wb") as pickle_file:
            pickle.dump("some_other_data", pickle_file)
        dat1 = open("test/pickle_reader_blob1.pkl", "rb").read()
        dat2 = open("test/pickle_reader_blob2.pkl", "rb").read()
        return [dat1, dat2]

    monkeypatch.setattr(reader,'download_blobs_as_strings',fake_blobs)

    reader_object, terminate = reader.run(data_object)

    assert not terminate

    assert "myreader" in reader_object.data_dict

    dat = reader_object.data_dict['myreader']

    assert "reader_data" in dat
    datlist = dat['reader_data']

    assert len(datlist) == 2

    assert "some_data" in datlist
    assert "some_other_data" in datlist

    files = ["test/pickle_reader_blob1.pkl", "test/pickle_reader_blob2.pkl"]
    for f in files:
        if os.path.exists(f):
            os.remove(f)

def test_run_pickle_2(monkeypatch):
    # returns 1 objects from pickle reader
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDillReader",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "deserializer": "pickle",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDeserializer(configuration, "myreader")

    data_object = DataObject(configuration)

    def fake_blobs():
        with open("test/pickle_reader_blob1.pkl", "wb") as pickle_file:
            pickle.dump("some_data", pickle_file)
        dat1 = open("test/pickle_reader_blob1.pkl", "rb").read()
        return [dat1]

    monkeypatch.setattr(reader,'download_blobs_as_strings',fake_blobs)

    reader_object, terminate = reader.run(data_object)

    assert not terminate

    assert "myreader" in reader_object.data_dict

    dat = reader_object.data_dict['myreader']

    assert "reader_data" in dat
    data = dat['reader_data']

    assert "some_data" == data

    files = ["test/pickle_reader_blob1.pkl"]
    for f in files:
        if os.path.exists(f):
            os.remove(f)

def test_run_other(monkeypatch):
    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "GcsDillReader",
                    "bucket_name": "test1",
                    "blob_name": "test2",
                    "deserializer": "other",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = GcsDeserializer(configuration, "myreader")

    data_object = DataObject(configuration)

    with pytest.raises(Exception, match=r"Unsupported"):
        reader.run(data_object)







"""Module with AbstractReader implementation, able to read from MySQL

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.sql_reader import AbstractSqlReader
from primrose.readers.mysql_helper import MySQLHelper

class MySQLReader(AbstractSqlReader):
    """Runs MySQL queries into pandas dataframes"""

    def get_connection(self):
        """return connection to MySQL DB

        Returns:
            connection to MySQL DB

        """
        return MySQLHelper.create_db_connection()



"""Module with AbstractNode implementation, able to read from local dill files

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""

import logging
import dill
import warnings

from primrose.base.reader import AbstractReader


class DillReader(AbstractReader):
    """Read a file from Gcs and un-dills it into memory"""

    DATA_KEY = 'reader_data'
    warnings.warn('Use Deserializer instead. DillReader will be deprecated in a future release.', DeprecationWarning)
    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the DillReader object

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            filename: local filename to be de-serialized

        Returns:
            set of necessary keys for the DillReader object

        """
        return set(['filename'])

    def run(self, data_object):
        """Read dill object(s) from local filesystem

        Args:
            data_object: DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        logging.info('Reading {} from local filesystem'.format(self.node_config['filename']))

        object = dill.load(open(self.node_config['filename'], 'rb'))

        data_object.add(self, object, key=DillReader.DATA_KEY)

        terminate = False

        return data_object, terminate




import pytest
import sys
import os
import pandas as pd
from primrose.configuration.configuration import Configuration
from primrose.readers.r_reader import RReader
from primrose.node_factory import NodeFactory
from primrose.data_object import DataObject
from primrose.base.node import AbstractNode
from primrose.data_object import DataObject, DataObjectResponseType


def check_rpy2():
    try:
        import rpy2
        return False
    except ImportError as e:
        return True

@pytest.mark.skipif(check_rpy2(), reason="primrose[R] is optional")
def test_run():
    config = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "RReader",
                    "dataset": "iris",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = RReader(configuration, "read_data")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    df = data_object.get('read_data', rtype=DataObjectResponseType.VALUE.value)
    assert df is not None
    assert df.shape == (150, 6)
    assert list(df.columns) == ["Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width", "Species", "row_names"]

@pytest.mark.skipif(check_rpy2(), reason="primrose[R] is optional")
def test_run2():
    config = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "RReader",
                    "dataset": "euro",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = RReader(configuration, "read_data")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    df = data_object.get('read_data', rtype=DataObjectResponseType.VALUE.value)
    assert df is not None
    assert df.shape == (11, 2)
    assert list(df.columns) == ["euro", "row_names"]




import pytest
import sys
import os
import pandas as pd
from primrose.configuration.configuration import Configuration
from primrose.writers.csv_writer import CsvWriter
from primrose.readers.csv_reader import CsvReader
from primrose.data_object import DataObject

@pytest.fixture()
def config():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["recipe_csv_writer"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    return config

def test_necessary_config(config):
    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    writer = CsvWriter(configuration, 'recipe_csv_writer')
    node_config = {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
    assert isinstance(writer.necessary_config(node_config), set)
    assert len(writer.necessary_config(node_config)) > 0

def test_init_ok(config):
    corpus = pd.read_csv("test/minimal.csv")

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    writer = CsvWriter(configuration, 'recipe_csv_writer')

    data_object = DataObject(configuration)
    requestor = CsvReader(configuration, 'csv_reader')
    data_object.add(requestor, key='test_data', data=corpus)


    c = configuration.config_for_instance('recipe_csv_writer') #configuration.sec .writer_config['recipe_csv_writer']
    filename = c['dir'] + os.path.sep + c['filename']

    #clean out test file location 
    if os.path.exists(filename):
        os.remove(filename)

    writer.run(data_object)

    assert os.path.exists(filename)

    df = pd.read_csv(filename)

    assert corpus.equals(df)

@pytest.fixture()
def kwargs_config():
    config = {
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["recipe_csv_writer"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "kwargs": {
                        "header": None,
                        "sep": ":"
                    }
                }
            }
        }
    }
    return config

def test_kwargs(kwargs_config):
    corpus = pd.read_csv("test/minimal.csv")

    configuration = Configuration(None, is_dict_config=True, dict_config=kwargs_config)
    writer = CsvWriter(configuration, 'recipe_csv_writer')

    data_object = DataObject(configuration)
    requestor = CsvReader(configuration, 'csv_reader')
    data_object.add(requestor, key='test_data', data=corpus)


    c = configuration.config_for_instance('recipe_csv_writer') #configuration.sec .writer_config['recipe_csv_writer']
    filename = c['dir'] + os.path.sep + c['filename']

    #clean out test file location
    if os.path.exists(filename):
        os.remove(filename)

    writer.run(data_object)

    assert os.path.exists(filename)

    df = pd.read_csv(filename,header=None, sep=':')

    assert df.shape == (2,2)


import pytest
from sklearn.tree import DecisionTreeClassifier
from primrose.configuration.configuration import Configuration
from primrose.readers.dill_reader import DillReader
from primrose.data_object import DataObject, DataObjectResponseType


def test_init_ok():
    config = {
        "implementation_config": {
            "reader_config": {
                "dill_reader": {
                    "class": "DillReader",
                    "filename": "test/tinymodel.dill",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DataObject(configuration)

    reader = DillReader(configuration, "dill_reader")
    data_object, terminate = reader.run(data_object)
    assert not terminate
    data = data_object.get('dill_reader', rtype=DataObjectResponseType.VALUE.value)

    assert data is not None
    assert set(data.keys()) == {'test', 'model'}

    node_config = {
                    "class": "DillReader",
                    "filename": "test/tinymodel.dill",
                    "destinations": []
                }

    assert isinstance(DillReader.necessary_config(node_config), set)
    assert len(DillReader.necessary_config(node_config)) > 0

    assert data['test'] == [1, 2, 3]
    assert isinstance(data['model'], DecisionTreeClassifier)



#########################################################################################
#
#
#
#
# *** IMPORTANT ***
#
# We highly recommend setting the environment variable PRIMROSE_EXT_NODE_PACKAGE to define
# the location of your custom node classes. Example: `export PRIMROSE_EXT_NODE_PACKAGE=src/mypackage`
#
# When this variable is set, primrose will automatically register any class
# present in the PRIMROSE_EXT_NODE_PACKAGE path which implements `base.node.AbstractNode`.
#
#
#
#
# Alternatively, if you would like to register classes yourself via `node_factory.NodeFactory`, you can follow the steps below.
#
# Overview
# ========
# There are three steps to registering your own classes:
#
# 1) you need to generate a script to run primrose from your own project. To do that
# run 
# 
#   primrose generate_script --destination path/to/myproject/
# 
# and it copies a run_primrose.py script to your project path
#
# 2) You need to register your own classes. This is the code in this script.
#
# 3) You need to reference this code in the run_primrose script.
#
# Step 1:
# ========
# run
#
#    primrose generate_script --destination path/to/myproject/
#
# which will create path/to/myproject/run_primrose.py
#
# Open up the file and see where it tells you to reference this code.
#
# Step2:
# ========
# Modify this file to register your own classes.
# As can be seen below, 
#  - you first need to import the NodeFactory singleton
#  - next, import your new classes
#  - finally, use NodeFactory to register those classes
#
# Step 3:
# ========
# *Importantly*, this factory registration has to occur *before* Configuration is instantiated
# in the run_primrose script. 
# To that end, we suggest putting this registration code below into something 
# like `src/__init__.py` in your project. Wherever you put it, you will need to reference 
# it in the run_primrose script.
#
# That is, if you put this code into `src/__init__.py`, you will need to add
#
#  from src.__init__ import *
#
# at the head of the run_primrose script. 
#
#########################################################################################

import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(filename)s %(funcName)s: %(message)s', level=logging.INFO)

from primrose.node_factory import NodeFactory

# Add your imports here
from src.yourpackage.awesome_reader import AwesomeReader
from src.yourpackage.awesome_model import AwesomeModel

NodeFactory().register_module_classes(__name__)




import pytest
from primrose.base.pipeline import AbstractPipeline
from primrose.base.pipeline import PipelineModeType
from primrose.configuration.configuration import Configuration
from primrose.data_object import DataObject
from primrose.readers.csv_reader import CsvReader
from primrose.base.transformer_sequence import TransformerSequence
from primrose.base.transformer import AbstractTransformer
from primrose.node_factory import NodeFactory
import pandas as pd
import logging
from testfixtures import LogCapture

def test_names():
    assert PipelineModeType.names() == ['FIT', 'FIT_TRANSFORM', 'TRANSFORM']

def test_values():
    assert PipelineModeType.values() == ['FIT', 'FIT_TRANSFORM', 'TRANSFORM']

def test_run():

    class TestPipeline(AbstractPipeline):
        def transform(self, data_object):
            logging.info("TRANSFORM CALLED")
            return data_object 

        def fit_transform(self, data_object):
            logging.info("FIT_TRANSFORM CALLED")
            return self.transform(data_object)

        @staticmethod
        def necessary_config(node_config): return set(['is_training'])

    NodeFactory().register("TestPipeline", TestPipeline)

    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mypipeline"]
                }
            },
            "pipeline_config": {
                "mypipeline":{
                    "class": "TestPipeline",
                    "is_training": True
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reference_file_path = "test/minimal.csv"
    corpus = pd.read_csv(reference_file_path)

    reader = CsvReader(configuration, "myreader")

    data_object = DataObject(configuration)
    data_object.add(reader, corpus)

    pipeline = TestPipeline(configuration, "mypipeline")

    with LogCapture() as l:
        pipeline.run(data_object)
    l.check(
        ('root', 'INFO', 'No upstream TransformerSequence found. Creating new TransformerSequence...'),
        ('root', 'INFO', 'FIT_TRANSFORM CALLED'),
        ('root', 'INFO', 'TRANSFORM CALLED')
    )

    data_object.add(reader, TransformerSequence(), "tsequence")
    with LogCapture() as l:
        pipeline.run(data_object)
    l.check(
        ('root', 'INFO', 'Upstream TransformerSequence found, initializing pipeline...'),
        ('root', 'INFO', 'FIT_TRANSFORM CALLED'),
        ('root', 'INFO', 'TRANSFORM CALLED')
    )

    config['implementation_config']['pipeline_config']['mypipeline']['is_training'] = False
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    reader = CsvReader(configuration, "myreader")
    data_object = DataObject(configuration)
    data_object.add(reader, corpus)
    pipeline = TestPipeline(configuration, "mypipeline")
    with LogCapture() as l:
        pipeline.run(data_object)
    l.check(
        ('root', 'INFO', 'No upstream TransformerSequence found. Creating new TransformerSequence...'),
        ('root', 'INFO', 'TRANSFORM CALLED')
    )

def test_execute_pipeline():
    class TestTransformer(AbstractTransformer):
        def fit(self, data):
            logging.info("Transfer FIT CALLED")

        def transform(self, data):
            logging.info("Transfer TRANSFORM CALLED")
            return data

        def fit_transform(self, data):
            logging.info("Transfer FIT_TRANSFORM CALLED")
            self.fit(data)
            return self.transform(data)

    class TestPipeline2(AbstractPipeline):
        def transform(self, data_object): return data_object 
        @staticmethod
        def necessary_config(node_config): return set(['is_training'])
    NodeFactory().register("TestPipeline2", TestPipeline2)

    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ["mypipeline"]
                }
            },
            "pipeline_config": {
                "mypipeline":{
                    "class": "TestPipeline",
                    "is_training": True
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reference_file_path = "test/minimal.csv"
    corpus = pd.read_csv(reference_file_path)

    reader = CsvReader(configuration, "myreader")

    data_object = DataObject(configuration)
    data_object.add(reader, corpus)

    sequence = TransformerSequence()
    sequence.add(TestTransformer())
    data_object.add(reader, sequence, "tsequence")


    pipeline = TestPipeline2(configuration, "mypipeline")

    with pytest.raises(Exception) as e:
         pipeline.execute_pipeline(corpus, PipelineModeType.FIT)
    assert "run() must be called to extract/create a TransformerSequence" in str(e)

    pipeline.run(data_object)

    with pytest.raises(Exception) as e:
         pipeline.execute_pipeline(corpus,"JUNK")
    assert "mode must be of type PipelineModeType Enum object." in str(e)

    with LogCapture() as l:
        pipeline.execute_pipeline(corpus, PipelineModeType.FIT)
    l.check(
        ('root', 'INFO', 'Transfer FIT CALLED')
    )

    with LogCapture() as l:
        pipeline.execute_pipeline(corpus, PipelineModeType.FIT_TRANSFORM)
    l.check(
        ('root', 'INFO', 'Transfer FIT_TRANSFORM CALLED'),
        ('root', 'INFO', 'Transfer FIT CALLED'),
        ('root', 'INFO', 'Transfer TRANSFORM CALLED')
    )


import pytest
import sys
import os
import pandas as pd
from primrose.configuration.configuration import Configuration
from primrose.pipelines.encode_train_test_split import EncodeTrainTestSplit
from primrose.data_object import DataObject, DataObjectResponseType
from primrose.readers.csv_reader import CsvReader


@pytest.fixture()
def configuration():
    return Configuration("test/hello_world_tennis.json")


@pytest.fixture
def data_obj_factory(configuration):

    def data_object_factory():
        df = pd.read_csv("test/tennis.csv")

        data_object = DataObject(configuration)

        csv_reader = CsvReader(configuration, "read_data")

        data_object.add(csv_reader, df)

        return data_object

    return data_object_factory


@pytest.fixture()
def pipeline_obj(configuration):
    return EncodeTrainTestSplit(configuration, "encode_and_split")


def test_concatenate_data(pipeline_obj, configuration):

    df1 = pd.read_csv("test/tennis.csv")
    df2 = pd.read_csv("test/tennis.csv")

    data_object = DataObject(configuration)
    csv_reader = CsvReader(configuration, "read_data")

    data_object.add(csv_reader, df1, 'query1')
    data_object.add(csv_reader, df2, 'query2')

    data_object, terminate = pipeline_obj.run(data_object)

    encoded_data = data_object.get('encode_and_split')['data_train']

    assert len(encoded_data) == 18


def test_encode_data(data_obj_factory, pipeline_obj):

    data_obj = data_obj_factory()

    data_object, terminate = pipeline_obj.run(data_obj)

    encoded_data = data_object.get('encode_and_split')['data_train']
    encoded_target = data_object.get('encode_and_split')['target_train']

    assert set(list(encoded_data.outlook.unique())) == set([0, 1, 2])
    assert set(list(encoded_data.temp.unique())) == set([0, 1, 2])
    assert set(list(encoded_data.humidity.unique())) == set([0, 1])
    assert set(list(encoded_data.windy.unique())) == set([True, False])
    assert set(list(encoded_target.unique())) == set([0, 1])

    assert pipeline_obj.transformer_sequence is not None
    assert list(pipeline_obj.first_transformer_in_sequence.target_encoder.inverse_transform([1, 0])) == ['yes', 'no']


def test_transform_after_fit(data_obj_factory, pipeline_obj):

    to_fit_object = data_obj_factory()
    data_obj = data_obj_factory()

    _data_object, _terminate = pipeline_obj.run(to_fit_object)

    data_object = pipeline_obj.transform(data_obj)

    transform_test_data_to_compare = data_object.get('encode_and_split')['data_test']

    assert len(transform_test_data_to_compare) == 14

    for col in transform_test_data_to_compare:
        assert transform_test_data_to_compare[col].dtype != object


def test_train_test_split(pipeline_obj):

    df = pd.read_csv("test/tennis.csv")
    train_data, test_data = pipeline_obj._train_test_split(df)

    assert len(train_data) == 9
    assert len(test_data) == 5
    assert not train_data.equals(test_data)




"""Abstract base class that performs TFIDF on some corpus

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from abc import abstractmethod
from primrose.base.model import AbstractModel
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from primrose.data_object import DataObjectResponseType
import pandas as pd
import logging

class AbstractSearchEngine(AbstractModel):
    """an abstract search engine"""

    def __init__(self, configuration, instance_name):
        """abstract search engine

        Args:
            configuration (Configuration): configuration object
            instance_name (str): name of the instance

        """
        super(AbstractSearchEngine, self).__init__(configuration, instance_name)
        self.ids = None
        self.docs = None
        self.tfidf = None
        self.term_document_matrix = None
        self.corpus_key = self.node_config.get('corpus_key','corpus')

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys for AbstractSearchEngine

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            id_key: key used in the corpus object for ids
            doc_key: key used in the corpus object for docs

        Returns:
            set of keys necessary to run AbstractSearchEngine

        """
        return set(['id_key', 'doc_key']).union(AbstractModel.necessary_config(node_config))

    def train_model(self, data_object):
        """train the model which means run fit_transform on a TFIDF model

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """

        upstream_data = data_object.get_upstream_data(self.instance_name,
            rtype=DataObjectResponseType.VALUE.value)

        if isinstance(upstream_data, pd.DataFrame):
            corpus = upstream_data

        # if a data dict was passed, loop through data objects searching for a corpus dataframe
        elif isinstance(upstream_data, dict):
            for data_key, data in upstream_data.items():

                # check that the data object is a dataframe, if not skip
                if not isinstance(data, pd.DataFrame):
                    logging.info("Upstream data from {}.{} not in necessary format, skipping".format(self.instance_name,
                                                                                                            data_key))
                    continue

                if data_key == self.corpus_key:
                    corpus = data
                else:
                    logging.info("{} key not found for data entry in upstream {} object, skipping".format(
                        self.corpus_key, data_key))
        else:
            raise Exception('Search Engine requires a corpus dataframe'.format())


        self.ids = list(corpus[self.node_config['id_key']])
        self.docs = list(corpus[self.node_config['doc_key']])
        self.tfidf = TfidfVectorizer(tokenizer=self.tokenize,stop_words=None,ngram_range=(1,1))
        self.term_document_matrix = self.tfidf.fit_transform(self.docs)

        return data_object

    def eval_model(self, data_object):
        """evaluate the model

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """
        # this is define as abstract in superclass
        # this code is essentially unreachable as it has to be overridden in concrete subclass
        return data_object

    def predict(self, data_object):
        """make predictions from this corpus, meaning create cosine similarity matrix

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """
        data_object = self.train_model(data_object)

        data_object.add(self, self.cosine_similarity_matrix())

        return data_object

    @abstractmethod
    def tokenize(self, s):
        """Given some string, tokenize it

        Args:
            s (str): input string

        Returns:
            list of tokens

        """
        # this code is essentially unreachable as it has to be overridden in concrete subclass
        pass # pragma: no cover

    def cosine_similarity_matrix(self):
        """compute the cosine similarities between all document pairs in the corpus

        Returns:
            matrix (numpy): square matrix of cosine similarities where index of matrix is index of corpus IDs

        """
        return cosine_similarity(self.term_document_matrix)



"""
    Module with tests for the success notification implementation
    author(s): Parul Laul (parul.laul@ww.com)

"""

import os
import unittest
import unittest.mock as mock

from primrose.data_object import DataObject
from primrose.node_factory import NodeFactory
from primrose.base.reader import AbstractReader
from primrose.configuration.configuration import Configuration
from primrose.notifications.success_notification import get_client_params, ClientNotification


config_dict = {
    "metadata": {
        "section_registry": [
        "initialize",
        "cleanup_config"
    ],
        "notify_on_error": {
            "client": "SlackClient",
            "channel": "some-channel",
            "member_id": None,
            "token": "some-token"
        }
    },
    "implementation_config": {
        "initialize": {
            "start_notification": {
                "class": "ClientNotification",
                "client": "SlackClient",
                "channel": "some-channel",
                "message": "starting job...",
                "member_id": "USomeUserID",
                "token": "some-token",
                "destinations": [
                    "notification"
                ]
            }
        },
        "cleanup_config": {
            "notification": {
                "class": "ClientNotification",
                "client": "SlackClient",
                "channel": "some-channel",
                "message": "TEST SUCCESS!",
                "member_id": "USomeUserID",
                "token": "some-token"
            }
        }
    }
}


config_dict_node_message = {
    "metadata": {
        "section_registry": [
        "reader_config",
        "cleanup_config"
    ],
        "notify_on_error": {
            "client": "SlackClient",
            "channel": "some-channel",
            "member_id": None,
            "token": "some-token"
        }
    },
    "implementation_config": {
        "reader_config": {
            "test_node": {
                "class": "SlackDataMock",
                "destinations": [
                    "node_notification"
                ]
            }
        },
        "cleanup_config": {
            "node_notification": {
                "class": "ClientNotification",
                "client": "SlackClient",
                "channel": "some-channel",
                "message": "TEST SUCCESS!",
                "member_id": "USomeUserID",
                "token": "some-token",
                "use_configuration_file_message" : False,
                "node_name" : "test_node",
                "message_key" : "test"
            }
        }
    }
}

class SlackDataMock(AbstractReader):
    def __init__(self, configuration, instance_name):
        super().__init__(configuration, instance_name)

    @staticmethod
    def necessary_config(node_config):
        return set()

    def run(self, data_object):

        data_object.add(self, "Node Success!", "test")

        return data_object

class TestClientNotification(unittest.TestCase):
    """Tests for success_notification.py"""


    def test_get_client_params(self):
        os.environ["SLACKCLIENT_CHANNEL"] = "test-channel"
        os.environ["SLACKCLIENT_MEMBER_ID"] = "test-member_id"
        os.environ["SLACKCLIENT_TOKEN"] = "test-token"

        params = {
            "client": "SlackClient",
            "message": "starting job..."
        }

        ans = get_client_params(params)
        expected = {
            "client": "SlackClient",
            "channel": "test-channel",
            "message": "starting job...",
            "member_id": "test-member_id",
            "token": "test-token"
        }
        self.assertDictEqual(ans, expected)


    def test_necessary_config(self):
        self.assertEqual(
            first=ClientNotification.necessary_config(node_config={}),
            second={'client', 'token'}
        )

    def test_run(self):

        path = 'primrose.notifications.success_notification.get_notification_client'
        with mock.patch(path) as get_client_mock:
            get_client_mock.return_value = mock.Mock()

            configuration = Configuration(None, is_dict_config=True, dict_config=config_dict)
            success_instance = ClientNotification(
                configuration=configuration,
                instance_name='notification'
            )
            success_instance.client = get_client_mock.return_value

            success_instance.run('some_data_object')

            success_instance.client.post_message.assert_called_once_with(message='TEST SUCCESS!')


    def test_run_node(self):

        path = "primrose.notifications.success_notification.get_notification_client"
        with mock.patch(path) as get_client_mock:
            get_client_mock.return_value = mock.Mock()

            NodeFactory().register("SlackDataMock", SlackDataMock)

            config = Configuration(None, is_dict_config=True, dict_config=config_dict_node_message)
            data_object = DataObject(config)

            reader = SlackDataMock(config, "test_node")
            data_object = reader.run(data_object)

            success_instance = ClientNotification(configuration=config, instance_name="node_notification",)
            success_instance.client = get_client_mock.return_value

            success_instance.run(data_object)

            success_instance.client.post_message.assert_called_once_with(message="Node Success!")



if __name__ == '__main__':
    unittest.main()



"""Module to run a basic clustering model

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import logging
from primrose.models.sklearn_model import SklearnModel

class SklearnClusterModel(SklearnModel):

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys

        Note:
            X (list): list of columns to use

            While you might expect `model` here, we do not need it when in predict or eval mode as the model is cached, only in train

        Returns:
            set of required keys

        """
        return SklearnModel.necessary_config(node_config)

    def fit_training_data(self):
        """fit training data to model"""
        self.model.fit(self.X_train)

    def _make_predictions(self, data_object):
        """Make predictions

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            nothing. Predictions stored in self.predictions

        """
        if hasattr(self, 'predictions') and self.predictions is not None:
            return

        if self.model is None:
            self.model = self.load_model(data_object)

        self.X_train, self.y_train, self.X_test, self.y_test = self._get_data(data_object)

        logging.info("Making predictions with model")
        to_predict = self.X_test

        if self.X_test is None:
            to_predict = self.X_train
            self.predictions = self.model.labels_
        else:
            self.predictions = self.model.predict(self.X_test)

    def predict(self, data_object, load_model=False, use_serial=False):
        """Predict y_test from X_test

        Args:
            data_object (DataObject): instance of DataObject

            load_model: load model object from gcs or not

        Returns: 
            data_object (DataObject): instance of DataObject

        """
        self._make_predictions(data_object)
        
        if self.X_test is None:
            data = self.X_train
            if self.y_train is not None:
                data['actual'] = self.y_train
        else:
            data = self.X_test
            if self.y_test is not None:
                data['actual'] = self.y_test

        data['predictions'] = self.predictions
        
        data_object.add(self, data)
        return data_object

    def get_scores(self):
        """get the scores for X_test
        
        Returns:
            returns a dictionary of scors
        
        """
        data = self.X_test
        if self.X_test is None:
            data = self.X_train
        return SklearnModel.evaluate_no_ground_truth_classifier_metrics(data, self.predictions)



"""Module with abstract model class to specify interface needed for future models

Author(s): 
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""

from abc import abstractmethod
from primrose.base.node import AbstractNode
from primrose.util import RunModes


class AbstractModel(AbstractNode):
    """Model class should be able to train, evaluate or predict"""

    def run(self, data_object):
        """run the model

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """

        mode = self.node_config['mode'].lower()
        assert mode in RunModes.values()

        if mode == RunModes.TRAIN.value:
            data_object = self.train_model(data_object)

        if mode in [RunModes.EVAL.value, RunModes.TRAIN.value]:
            data_object = self.eval_model(data_object)

        data_object = self.predict(data_object)
        terminate = False
        return data_object, terminate

    @staticmethod
    @abstractmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            After adding this list, validation automatically occurs before instantiation in the pipeline factory.

        Returns:
            set of keys necessary to run implementation

        """
        return set(['mode']) # pragma: no cover

    @abstractmethod
    def train_model(self, data_object):
        """Train an internal model attribute, then save the trained model with the save_model method

            Using the training features in feature_df (pandas dataframe object) and the target_variable (pandas series),
            cross-validation or other training scripts will happen in theis method, according to the parameters
            in parameters

        Args:
            data_object (DataObject): instance of DataObject

        Returns: 
            data_object (DataObject): instance of DataObject

        """
        pass # pragma: no cover

    @abstractmethod
    def eval_model(self, data_object):
        """Evaluate a previously trained model performance using labeled data

            Method should be able to load a serialized model if necessary (load_model method), or work from a model
            trained in-scope with the train_model method. This method should calculate and store some aspect of
            model error into attributes that are serialized with the save_model method

        Args:
            data_object (DataObject): instance of DataObject

        Returns: 
            data_object (DataObject): instance of DataObject

        """
        pass # pragma: no cover

    @abstractmethod
    def predict(self, data_object):
        """Make predictions using a pre-trained model on the features in the feature_df dataframe

            Using a pre-trained model (either from an in-scope run of train_model or a call to the load_model method)
            this method should append predictions to each row of features in the input, feature_df. This method may
            also add feature importance columns, for importance analysis

        Args:
            data_object: DataObject instance

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        pass # pragma: no cover




import pytest
import os
from primrose.base.sql_reader import AbstractSqlReader
from primrose.configuration.configuration import Configuration
from primrose.readers.sqlite_reader import SQLiteReader
import sqlite3

def test__substitute_query():

    query = """SELECT * FROM TABLE where field={x} and otherfield='{y}'"""

    filename = "test/test_substitute.sql"
    if os.path.exists(filename):
        os.remove(filename)

    with open(filename, 'w') as f:
        f.write(query)

    individual_query_json = {
        "query": "test/test_substitute.sql",
        "parameters": {
            "x": 3,
            "y": "somestring"
        }
    }
    class TestSqlReader(AbstractSqlReader):
        pass   
    
    transformed_query = TestSqlReader._substitute_query(individual_query_json)

    assert transformed_query == "SELECT * FROM TABLE where field=3 and otherfield='somestring'"

    if os.path.exists(filename):
        os.remove(filename)

def test__substitute_query2():
    query_00 = """SELECT * FROM TABLE where field={x} and otherfield='{y}'"""
    query_01 = """SELECT * FROM TABLE where field={a} and otherfield='{b}'"""
    query_02 = """
    {% include 'test/test_substitute_00.sql' %}
    UNION ALL
    {% include 'test/test_substitute_01.sql' %}
    """
    queries = [query_00, query_01, query_02]
    
    filenames = []
    for i in range(3):
        filename = "test/test_substitute_{:02d}.sql".format(i)
        filenames.append(filename)
        if os.path.exists(filename):
            os.remove(filename)

        with open(filename, 'w') as f:
            f.write(queries[i])

    individual_query_json = {
        "query": "test/test_substitute_02.sql",
        "parameters": {
            "a": 2,
            "b": "somestring2",
            "x": 3,
            "y": "somestring"
        }
    }
    class TestSqlReader(AbstractSqlReader):
        pass   
    
    transformed_query = TestSqlReader._substitute_query(individual_query_json)

    expected_query = """
    SELECT * FROM TABLE where field=3 and otherfield='somestring'
    UNION ALL
    SELECT * FROM TABLE where field=2 and otherfield='somestring2'
    """
    print(transformed_query)

    assert transformed_query == expected_query

    for filename in filenames:
        if os.path.exists(filename):
            os.remove(filename)

def test__generate_queries():

    config = {
        "implementation_config": {
            "reader_config": {
                "myreader": {
                    "class": "SQLiteReader",
                    "filename": "test/test_abstract_sqlite.db",
                    "destinations": [],
                    "query_json": [
                        {
                            "query": "test/test_substitute1.sql",
                            "parameters": {
                                "firstname": "joe",
                                "lastname": "doe2"
                            }
                        },
                        {
                            "query": "test/test_substitute2.sql",
                            "parameters": {
                                "firstname": 'mary',
                                "lastname": "poppins"
                            }
                        }
                    ]
                }
            }
        }
    }

    filenames = ["test/test_substitute1.sql", "test/test_substitute2.sql"]

    for filename in filenames:

        query = """SELECT * FROM test where firstname='{firstname}' and lastname='{lastname}'"""

        if os.path.exists(filename):
            os.remove(filename)

        with open(filename, 'w') as f:
            f.write(query)

    db_filename = "test/test_abstract_sqlite.db"
    if os.path.exists(db_filename):
        os.remove(db_filename)

    conn = sqlite3.connect(db_filename)
    c = conn.cursor()
    c.execute("create table test(firstname text, lastname text);")
    c.execute("insert into test(firstname, lastname) values('joe', 'doe'), ('joe', 'doe2'), ('mary','poppins');")
    conn.commit()
    conn.close()

    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    reader = SQLiteReader(configuration, "myreader")
    queries = [q for q in reader._generate_queries()]

    assert len(queries) == 2
    assert queries[0] == "SELECT * FROM test where firstname='joe' and lastname='doe2'"
    assert queries[1] == "SELECT * FROM test where firstname='mary' and lastname='poppins'"

    for filename in filenames:
        if os.path.exists(filename):
            os.remove(filename)
    
    if os.path.exists(db_filename):
        os.remove(db_filename)


"""Module to run a basic decision tree model

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""
import logging
import datetime
from sklearn.base import is_classifier
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import f1_score
from sklearn.metrics import recall_score
from sklearn.metrics import precision_score
from sklearn.metrics import accuracy_score
from sklearn.metrics import auc
from primrose.readers.dill_reader import DillReader
from primrose.models.sklearn_model import SklearnModel

class SklearnClassifierModel(SklearnModel):

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys

        Args:
            node_config (dict): set of parameters / attributes for the node

        Notes:
            model_parameters (dict): parameters that mirror the sklearn kwargs for the user's model
            mode: train, eval or predict (see AbstractModel)
            sklearn_classifier_name: sklearn submodule and model name (submodule.model_name) of the user's model
            grid_search_scoring: scoring function name from sklearn CV docs
            cv_folds: number of CV folds

        Returns:
            set of required keys

        """
        return set(["model_parameters", "mode", "sklearn_classifier_name", "grid_search_scoring", "cv_folds"])

    def train_model(self, data_object):
        """train the model using CV, according to user specified options

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            Nothing

        """
        X_train, y_train, X_test, y_test = self._get_data(data_object)

        logging.info("Fitting model")

        self.model = SklearnModel._instantiate_model(self.node_config['sklearn_classifier_name'], args=None)

        self.model = GridSearchCV(self.model,
                                self.node_config['model_parameters'],
                                n_jobs=-1,
                                scoring=self.node_config['grid_search_scoring'],
                                verbose=2,
                                cv=self.node_config['cv_folds'],
                                refit=True)

        self.model.fit(X_train, y_train)

        data_object.add(self, self.model.best_estimator_, 'model')

        return data_object

    def eval_model(self, data_object):
        """Evaluate model perfomance on a labeled testing dataset
        
        Returns:
            data_object (DataObject): instance of DataObject
        
        """

        if self.model is None:
            self.model = self.load_model(data_object)

        logging.info("Evaluating model performance on testing data")

        X_train, y_train, X_test, y_test = self._get_data(data_object)

        model_predictions = self.model.predict(X_test)
        model_probability = self.model.predict_proba(X_test)

        model_predictions = model_predictions.astype(int)

        model_f1 = f1_score(y_test, model_predictions, labels=None, pos_label=1)
        recall = recall_score(y_test, model_predictions, labels=None, pos_label=1)
        prec = precision_score(y_test, model_predictions, labels=None, pos_label=1)
        accuracy = accuracy_score(y_test, model_predictions)

        logging.info('positive class fraction: {}'.format(float(sum(y_test)) / len(y_test)))
        logging.info('positive class predicted fraction: {}'.format(float(sum(model_predictions)) / len(y_test)))
        logging.info('f1: {}, recall: {}, precision: {}, accuracy: {}'.format(model_f1, recall, prec, accuracy))
        roc_auc = self._get_roc_score(y_test, model_probability)
        logging.info('AUC: {}'.format(roc_auc))

        # write performance and timing information to scores attribute
        self.scores['auc'] = roc_auc
        self.scores['f1'] = model_f1
        self.scores['recall'] = recall  # Recall=TP/(TP+FN)
        self.scores['precision'] = prec  # Precison=TP/(TP+FP)
        self.scores['positive_class_fraction'] = float(sum(y_test)) / len(y_test)
        self.scores['predicted_class_fraction'] = float(sum(model_predictions)) / len(y_test)
        self.scores['eval_time'] = datetime.datetime.now()

        return data_object

    def predict(self, data_object):
        """Make distance-based predictions using the prebuilt matrix

        Args:
            data_object: DataObject instance
            load_model: load model object from gcs or not

        Returns: 
            data_object with prediction data added

        """

        if self.model is None:
            self.model = self.load_model(data_object)

        X_train, y_train, X_test, y_test = self._get_data(data_object)

        predictions = self.model.predict(X_test)

        # get the upstream target_encoder if it exists
        data = data_object.get_filtered_upstream_data(self.instance_name, filter_for_key='target_encoder')

        if 'target_encoder' in data:
            logging.info("Reversing label encoding")
            predictions = data['target_encoder'].inverse_transform(predictions)

        # get original data frame and tack on column of predictions
        data_out = X_test
        data_out['predictions'] = predictions

        data_object.add(self, data_out, 'predictions')

        return data_object



"""
    Module to establish connection to slack app and post to channel
    Author(s): Parul Laul (parul.laul@ww.com)

"""

import importlib
import logging
import os

import slack
from slack.errors import SlackApiError

from primrose.notifications.abstract_notification import AbstractNotification


class SlackClient(AbstractNotification):
    """Module to establish connection to slack app and post to channel"""

    def __init__(self, channel, token, member_id=None):
        """Outputs messages in a specified slack channel to
        notify user of the status of the DAG after completion.

        Args:
            channel (`str`): Channel to post slack messages
            member_id (`str`): Slack member_id, defaults to None.
                Used if user would like to be pinged with the message output.

        Returns:
            None

        """
        self.channel = channel
        self.member_id = member_id  # this is not userid. See https://medium.com/@moshfeu/how-to-find-my-member-id-in-slack-workspace-d4bba942e38c
        self.client = slack.WebClient(token=token)

    def post_message(self, message: str):
        """Calls slack client api to post message in channel

        Args:
            message (`str`): message to be posted
        Returns:
            slack.web.slack_response.SlackResponse, and message is posted to slack.

        """

        if self.member_id:
            message += f"\n <@{self.member_id}>"

        # post response
        try:
            return self.client.chat_postMessage(channel=self.channel, text=message)

        except SlackApiError as error:
            msg = "Something went wrong when attempting to send message to slack"
            logging.error(msg=msg, exc_info=error)


def get_notification_client(params: dict):

    """Get client given configuration parameters

    Args:
        params (`dict`): parameters for client instantiation
        (example. {'client': 'SlackClient', 'channel': 'my_channel', 'token': 'some-token'})

    Returns:
        instantiated client object

    """
    exclude = [
        "class",
        "client",
        "message",
        "destinations",
        "use_configuration_file_message",
        "node_name",
        "message_key",
    ]
    client_params = {k: v for k, v in params.items() if k not in exclude}

    # instantiate client
    module = importlib.import_module(__loader__.name)
    try:
        client = getattr(module, params["client"])

        return client(**client_params)

    except AttributeError:
        msg = "Are you sure {} is in {}?".format(params["client"], module)
        logging.exception(msg)



"""Singleton Factory where one can register traversers for instantiation

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""

import logging
from primrose.dag.config_layer_traverser import ConfigLayerTraverser
from primrose.dag.depth_first_traverser import DepthFirstTraverser

class TraverserFactory:
    """Singleton Factory where one can register traversers for instantiation"""

    instance = None

    def __init__(self):
        """instantiate the factory but as a singleton. The guard raails are here
        """
        # where the magic happens, only one instance allowed:
        if not TraverserFactory.instance:
            TraverserFactory.instance = TraverserFactory.__HiddenFactory()

    def __getattr__(self, name):
        """getattr with instance name

        Returns:
            gettattr

        """
        return getattr(self.instance, name)

    class __HiddenFactory():
        """actual factory where registry and instantiation happens"""

        
        def __init__(self):
            """instantiate the HiddenFactory"""
            self.name_dict = {}
            self.register('ConfigLayerTraverser', ConfigLayerTraverser)
            self.register('DepthFirstTraverser', DepthFirstTraverser)
            self.DEFAULT_TRAVERSER = 'ConfigLayerTraverser'

        def register(self, key, class_obj, raise_on_overwrite=False):
            """Registering class_obj with key

            Args:
                key (str): key such as class name, e.g. 'ConfigLayerTraverser'
                class_obj (class obj), e.g. ConfigLayerTraverser

            Returns:
                nothing. Side effect is to register the class

            """
            if raise_on_overwrite and key in self.name_dict:
                raise Error("Traverser already exist with the key " + key)
            self.name_dict[key] = class_obj
            logging.debug("Registered %s : %s", key, class_obj)

        def default_traverser(self, configuration):
            """instantiate the default traverser class

            Args:
                configuration (Configuration): Configuration instance

            Returns:
                Traverser instance

            """
            return self.instantiate(self.DEFAULT_TRAVERSER, configuration)

        def instantiate(self, class_name, configuration):
            '''instantiate instances of rule, given name of rule class

            Args:
                class_name (str): name of the class
                configuration (Configuration): Configuration instance

            Returns:
                instance (Traverser): instance of a traverser

            '''
            return self.name_dict[class_name](configuration)



"""Simple success node: log a message at specified log level

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import logging
from primrose.base.success import AbstractSuccess

class LoggingSuccess(AbstractSuccess):
    """simple success node: log a message"""

    @staticmethod
    def necessary_config(node_config):
        """Return a list of necessary configuration keys within the implementation

        Args:
            node_config (dict): set of parameters / attributes for the node

        Note:
            msg: message you want logged
            level: one of 'DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'

        Returns:
            set of keys necessary to run implementation

        """
        return set(['msg', 'level']) 

    def run(self, data_object):
        """Signal success by logging a message at specified log level

        Args:
            data_object (DataObject): DataObject instance

        Returns:
            nothing. Side effect is to signal success via logging

        """
        msg = self.node_config['msg']

        level = str(self.node_config['level']).upper()

        # check whether valid level: will throw KeyError if level not recognized
        logging._nameToLevel[level]

        level = logging.getLevelName(level)

        logging.getLogger("")._log(level, msg, None, None)

        terminate = False

        return data_object, terminate



'''A module that reads something from somewhere

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

'''
from primrose.base.node import AbstractNode

class AbstractReader(AbstractNode):
    """read some data from somewhere"""
    pass



"""Module to plot scatter plot of clusters

Author(s): 
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.node import AbstractNode
from primrose.data_object import DataObjectResponseType
import matplotlib.pyplot as plt
import logging

class ClusterPlotter(AbstractNode):
    """Plot clusters on a 2D scatter plot"""

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the ClusterPlotter object

        Note:
            id_col (str): name of column with cluster IDs
            filename (str): name of the file

        Returns: 
            set of necessary keys for the CsvReader object

        """
        return set(['id_col', 'filename'])

    def run(self, data_object):
        """Create a PNG image of the clustered data

        Note:
            saves image to file specified in filename key
            optional key: `title` which is title of plot

        Returns:
            data_object (DataObject): DataObject instance. Here, it is unmodified
            terminate (bool): should we terminate the DAG? true or false. Terminate if empty data

        """
        dat = data_object.get_upstream_data(self.instance_name, pop_data=False, rtype=DataObjectResponseType.KEY_VALUE.value)
        X = dat['data']
        cluster_ids = X[self.node_config['id_col']]
        centers = X.groupby(self.node_config['id_col']).mean()

        plt.scatter(X.iloc[:,0],X.iloc[:,1], c=cluster_ids, s=50, cmap='viridis')
        plt.scatter(centers.iloc[:,0], centers.iloc[:,1], c='black', s=200, alpha=0.5)

        if self.node_config['title']:
            plt.title(self.node_config['title'])

        plt.savefig(self.node_config['filename'])
        logging.info("Saved cluster plot to " + self.node_config['filename'])

        terminate = X.empty
        return data_object, terminate



"""primrose wrapper around sklearn preprocessor

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import logging
import pandas as pd

from primrose.base.transformer import AbstractTransformer

class SklearnPreprocessingTransformer(AbstractTransformer):

    def __init__(self, preprocessor, columns):
        """initialize the proprocessor

        Args:
            preprocessor (SKlearn preprocessor): preprocesor from Sklearn
            columns (list of str): list of columns

        """
        #Ideally, would check that this is a preprocessor but there is no good class to check against
        self.preprocessor = preprocessor
        self.columns = columns

    def fit(self, data):
        """User implements fit operation on a single data element from a data_object
        
        Args:
            data (object): some data

        Returns:
            data

        """
        if isinstance(data, pd.DataFrame):
            if self.columns:
                self.preprocessor.fit(data[self.columns].values)
            else:
                self.preprocessor.fit(data.values)
        else:
            self.preprocessor.fit(data)

    def transform(self, data):
        """User implements internal transform function which operates on a single data element from a data_object
        
        Args:
            data (object): input data

        Returns:
            data, transformed

        """
        if isinstance(data, pd.DataFrame):
            if self.columns:
                scaled_features_df = data.copy()

                # transform select columns
                scaled_features = self.preprocessor.transform(data[self.columns].values)
                for idx, colname in enumerate(self.columns):
                    scaled_features_df[colname] = scaled_features[:,idx]

            else:
                scaled_features = self.preprocessor.transform(data.values)
                try:
                    scaled_features_df = pd.DataFrame(scaled_features, index=data.index, columns=data.columns)
                except ValueError:
                    logging.info(f'{self.preprocessor.__class__.__name__} instance changed the number of columns. Returning raw values')
                    return pd.DataFrame(scaled_features)
            return scaled_features_df

        return self.preprocessor.transform(data)

    def fit_transform(self, data):
        """fit then transform data

        Args:
            data (data): input data

        Returns:
            data, transformed

        """
        self.fit(data)
        return self.transform(data)


"""Module with AbstractReader implementation, able to read from PostgreSQL

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.sql_reader import AbstractSqlReader
from primrose.readers.postgres_helper import PostgresHelper

class PostgresReader(AbstractSqlReader):
    """Runs PostgreSQL queries into pandas dataframes"""

    def get_connection(self):
        """return connection to PostgreSQL DB

        Returns:
            connection to PostgreSQL DB

        """
        return PostgresHelper.create_db_connection()




import pytest
from primrose.configuration.configuration import Configuration
from primrose.dag.depth_first_traverser import DepthFirstTraverser

def test_traversal_list():
    config = {
        "implementation_config":{
            "reader_config":{
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['csv_writer']
                },
                "csv_reader2": {
                    "class": "CsvReader",
                    "filename": "test/minimal.csv",
                    "destinations": ['successlogger']
                }
            },
            "writer_config": {
                "csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv",
                    "destinations": ["successlogger"]
                }
            },
            "cleanup_config": {
                "successlogger": {
                    "class": "LoggingSuccess",
                    "msg": "woohoo, all done!",
                    "level": "INFO",
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    traverser = DepthFirstTraverser(configuration)
    sequence = traverser.traversal_list()
    print(sequence)
    assert sequence == ['csv_reader2', 'csv_reader', 'csv_writer', 'successlogger']




import pytest
from primrose.util import RunModes

def test_values():
    values = RunModes.values()
    assert len(values) == 3
    assert set(values) == set(['train', 'predict' ,'eval'])

def test_names():
    names = RunModes.names()
    assert len(names) == 3
    assert set(names) == set(['TRAIN', 'PREDICT' ,'EVAL'])


"""Run the DAG: gets list of nodes to traverse then calls run(data_object) on each

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.data_object import DataObject
import networkx as nx
import os
import logging
import collections
import traceback
from primrose.node_factory import NodeFactory
from primrose.configuration.configuration import OperationType
from primrose.notification_utils import get_notification_client
from primrose.dag.traverser_factory import TraverserFactory
from primrose.base.conditional_path_node import AbstractConditionalPath


class DagRunner():
    """class that runs the DAG: gets the list of nodes to traverse and then asks them to run"""

    def __init__(self, configuration):
        """instantiate the DagRunner

        Args:
            configuration (Configuration): configuration object defined in primrose/Configuration with validated inputs
                from the result of necessary_config, all inputs are described in that method

            instance_name (str): how the code knows where it is from

        """
        self.configuration = configuration
        self.dag = self.configuration.dag

        if configuration.config_metadata and 'traverser' in configuration.config_metadata:
            self.dag_traverser = TraverserFactory().instantiate(configuration.config_metadata['traverser'], configuration)
        else:
            self.dag_traverser = TraverserFactory().default_traverser(configuration)

        logging.info("Traverser is of class %s", self.dag_traverser.__class__)

    def create_data_object(self):
        """restore data_object from cache

        Returns:
            data_object (DataObject)

        """
        if self.configuration.config_metadata and 'data_object' in self.configuration.config_metadata:

            cfg = self.configuration.config_metadata['data_object']

            if "read_from_cache" in cfg and cfg['read_from_cache']:
                # we can assume that read_filename exists due to configuration checks
                filename = cfg['read_filename']
                assert os.path.exists(filename)

                logging.info("Reading DataObject from cache " + filename)
                return DataObject.read_from_cache(filename)

        data_object = DataObject(self.configuration)

        return data_object

    def cache_data_object(self, data_object):
        """cache the data object

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            whether it was cached (bool)

        """

        if self.configuration.config_metadata and 'data_object' in self.configuration.config_metadata:

            cfg = self.configuration.config_metadata['data_object']

            if "write_to_cache" in cfg and cfg['write_to_cache']:

                assert "write_filename" in cfg
                filename = cfg["write_filename"]
                data_object.write_to_cache(filename)
                return True

        return False

    def initial_check_sequence(self, sequence):
        """Some checks on the incoming sequence

        Args:
            sequence (list): list of nodes to run in given order

        Returns:
            nothing

        Raises:
            Exception if there are dupes in the sequence, or if nodes are not in config, or we have nodes from other sections.
            The latter can happen if we mix up nodes from sections. That is, suppose we have section1 (1 node) and section 2 (2 nodes) and
            we want to run section2 and then section1 and we receive sequence [section2_node1, section1_node1, section2_node2], it will
            complain about the partition [section2_node1, section1_node1] [section2_node2] as they are mixed from sections.

        """
        dupes = [item for item, count in collections.Counter(sequence).items() if count > 1]

        if len(dupes) > 0:
            raise Exception("You have duplicate nodes from traverser! " + str(dupes))

        # explictly check that each of these is a known node in config, raising exception if not
        [self.configuration.config_for_instance(instance_name) for instance_name in sequence]

    def filter_sequence(self, sequence):
        """The user may have specified some subset of sections to run in metadata.section_run
        Let's assume we can't trust traverser to limit themselves to those sections, so here
        we limit the sequence, if necessary

        Args:
            sequence (list): list of nodes to run in given order

        Returns:
            sequence (list): complete or subset of input sequence

        Raises:
            Exception if there are dupes in the sequence, or if nodes are not in config, or we have nodes from other sections.
            The latter can happen if we mix up nodes from sections. That is, suppose we have section1 (1 node) and section 2 (2 nodes) and
            we want to run section2 and then section1 and we receive sequence [section2_node1, section1_node1, section2_node2], it will
            complain about the partition [section2_node1, section1_node1] [section2_node2] as they are mixed from sections.

        """
        self. initial_check_sequence(sequence)

        # it might be that traverser provides a sequence of all config nodes and we only want to run those in section2
        # Thus, first, let's only consider the nodes in that section
        nodes_to_run = set()
        sections, source = self.configuration.sections_in_order()

        logging.info("Taking nodes to run from " + source)

        for section_name in sections:
            nodes_to_run = nodes_to_run.union(set(self.configuration.config[section_name].keys()))
        sequence = [n for n in sequence if n in nodes_to_run]

        if not self.dag_traverser.run_section_by_section():
            self.check_for_upstream(sequence)
            return sequence

        filtered_sequence = []

        for section_name in sections:

            #get a set of all nodes in this section
            this_section = set(self.configuration.config[section_name].keys())

            n = len(this_section)

            # do we have enough nodes in sequence to compare against?
            if len(sequence) >= n:

                subset = sequence[:n]

                #logic: the set of first n nodes of sequence should match this set of this_section
                if this_section == set(subset):

                    #these nodes passed the checks
                    filtered_sequence.extend(sequence[:n])

                    #pop off the first n nodes and repeat until we've gone through each section
                    sequence = sequence[n:]
                else:
                    # We can't get here if there is only 1 section to run as we raise on dupes and importantly, we now filter on sections earlier.
                    # This means that we cannot pull in nodes from a later section. That is, suppose the sequence was
                    # [section2_node1, section1_node1, section2_node2] and we wanted to run section 2 only (which has two nodes).
                    # We wouldn't grab first two nodes of sequence [section2_node1, section1_node1] which is what code below was
                    # meant to detect, but instead we would get [section2_node1, section2_node2] from the section filter.
                    # So, it would either match exactly or we would run out of nodes.
                    #
                    # For 2 or more nodes, however, it is different. If we wanted to run section 2 and then section 1, we would
                    # grab first 2 nodes [section2_node1, section1_node1] and fall into this section.
                    msg = "Traverser is mismatched with section " + section_name + "."
                    msg += " Expecting set " + str(sorted(this_section)) + ".\n"
                    msg += " Received list " + str(sorted(subset))
                    raise Exception(msg)
            else:
                raise Exception("Ran out of nodes for section " + section_name + ". Only received " + str(sequence))

        self.check_for_upstream(filtered_sequence)

        return filtered_sequence

    def check_for_upstream(self, sequence):
        """check for any upstream paths with the input sequence.
        That is, suppose we had a reader flowing to writer. It would not make sense
        to run writer before the reader.

        Args:
            sequence (list): list of node names

        Raises:
            Exception if any upstream paths found

        """

        for idx_from in range(len(sequence)):
            for idx_to in range(len(sequence)):
                if idx_from > idx_to:
                    if self.configuration.dag.paths(sequence[idx_from],  sequence[idx_to]):
                        msg = "Upstream path found, from %s to %s" % (sequence[idx_from], sequence[idx_to])
                        raise Exception(msg)
        return False

    def run(self, dry_run=False):
        """run the whole DAG. Optonally, you can call dry_run=True
            which will log what would be run and in what order
            but not actually run it

        Args:
            dry_run: Boolean. Want to do a dry run?

        Returns:
            data_object: DataObject instance
            node (Node): last node run

        """
        data_object = self.create_data_object()

        candidate_sequence = self.dag_traverser.traversal_list()

        sequence = self.filter_sequence(candidate_sequence)

        if len(candidate_sequence) > len(sequence):
            logging.info("Sequence of nodes to be run: %s", sequence)

        pruned_nodes = set()

        if (self.configuration.config_metadata and
            'notify_on_error' in self.configuration.config_metadata):
            try:
                params = self.configuration.config_metadata['notify_on_error']
                client = get_notification_client(params)

            except Exception as error:
                msg = (
                    'Error trying to instantiate notification client.'
                    'Check class name and parameters"'
                )
                logging.error(error)
                raise(msg)
        else:
            client = None

        for i, node in enumerate(sequence):

            if node in pruned_nodes:
                logging.info("Skipping pruned node " + node)
                continue

            section = self.dag.node_map[node]
            class_name = self.configuration.nodename_to_classname[node]

            if dry_run:
                logging.info("DRY RUN %s: would run node %s of type %s and class %s", i, node, section, class_name)
                continue
            else:
                logging.info("received node %s of type %s and class %s", node, section, class_name)

            try:
                node_instance = NodeFactory().instantiate(class_name, self.configuration, node)
            except Exception as e:
                msg = "Issue instantiating %s and class %s" % (node, class_name)
                logging.error(msg)
                if client:
                    client.post_message(msg)
                raise Exception(msg)

            try:
                data_object, terminate = node_instance.run(data_object)

                if isinstance(node_instance, AbstractConditionalPath):
                    to_prune = node_instance.all_nodes_to_prune()
                    if to_prune:
                        pruned_nodes.update(to_prune)

            except Exception as e:
                msg = "Issue with %s" % node
                logging.error(msg)
                if client:
                    client.post_message(f"{msg}\n{traceback.format_exc()}")
                raise e

            if terminate:
                logging.info("Terminating early due to signal from %s", node)
                if client:
                    client.post_message(msg)
                break

        self.cache_data_object(data_object)

        logging.info("All done. Bye bye!")
        return data_object



"""MySQL helper

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import mysql.connector
from primrose.readers.database_helper import get_env_val

class MySQLHelper():
    """"some utility methods for connecting to MySQL"""

    @staticmethod
    def extract_mysql_credentials():
        """extract MySQL credentials from config

        Returns:
            (tuple): tuple containing: \
                host (str): host \
                port (int): port \
                database (str): database name \
                username (str): username \
                password (str): password \

        """
        host = get_env_val("MYSQL_HOST")
        port = int(get_env_val("MYSQL_PORT"))
        database = get_env_val("MYSQL_DB")
        username = get_env_val("MYSQL_USER")
        password = get_env_val("MYSQL_PASS")
        return (host, port, username, password, database)

    @staticmethod
    def create_db_connection():
        '''authenticate with MySQL database

        Returns:
            db (connection object): MySQL db object

        '''

        host, port, username, password, database = MySQLHelper.extract_mysql_credentials()
        conn = mysql.connector.connect(database=database, user=username, passwd=password, host=host, port=port, auth_plugin='mysql_native_password')
        return conn



"""Abstract base class for a clean up node, such as signalling success

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""

from abc import abstractmethod
from primrose.base.node import AbstractNode

class AbstractSuccess(AbstractNode):
    """ Ability to cleanup, such as notify success"""
    pass



"""
    Abstract class to notify upon exception

    Author(s):
        Parul Laul (parul.laul@ww.com)

"""


from abc import ABC, abstractmethod


class AbstractNotification(ABC):

    @abstractmethod
    def post_message(self, message):
        """
        Sends message to the notification client.

        Args:
            message (str): Message to be sent

        """
        pass


"""Module to read canned datasets from sklearn

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.sql_reader import AbstractReader
from sklearn.datasets import *
import pandas as pd

class SklearnDatasetReader(AbstractReader):
    """Read data from sklearn.dataset into a pandas dataframe"""

    @staticmethod
    def necessary_config(node_config):
        """Returns the necessary configuration keys for the SklearnDatasetReader object

        Note:
            dataset (str): name of supported sklearn.dataset. One of "iris", "boston", "diabetes", "breast_cancer", "linnerud", "wine"
            
        Returns: 
            set of necessary keys for the SklearnDatasetReader object

        """
        return set(['dataset'])

    def run(self, data_object):
        """Read sklearn dataset to a pandas dataframe

        Returns:
            data_object (DataObject): DataObject instance
            terminate (bool): should we terminate the DAG? true or false

        """
        dataset = self.node_config['dataset']

        function_map = {
            "iris": load_iris,
            "boston": load_boston,
            "diabetes": load_diabetes,
            "breast_cancer": load_breast_cancer,
            "linnerud": load_linnerud,
            "wine": load_wine
        }

        if not dataset in function_map.keys():
            raise Exception("Dataset not supported " + dataset)

        data = function_map[dataset]()

        df = pd.DataFrame(data['data'], columns = data['feature_names'])
        df['target'] = data['target']

        data_object.add(self, df)

        terminate = False
        return data_object, terminate




import pytest
from primrose.configuration.configuration import Configuration
from primrose.dag_runner import DagRunner
import os
import math

def test_run():
    config = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "SklearnDatasetReader",
                    "dataset": "iris",
                    "destinations": [
                        "train_test_split"
                    ]
                }
            },
            "pipeline_config": {
                "train_test_split": {
                    "class": "TrainTestSplit",
                    "features": ["sepal length (cm)", "petal length (cm)", "petal width (cm)"],
                    "target_variable": "sepal width (cm)",
                    "training_fraction": 0.65,
                    "is_training": True,
                    "seed": 42,
                    "destinations": [
                        "regression_model"
                    ]
                }
            },
            "model_config": {
                "regression_model":{
                    "class": "SklearnRegressionModel",
                    "mode": "train",
                    "model": {"class": "linear_model.LinearRegression"},
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    data_object = DagRunner(configuration).run()
    scores = data_object.data_dict['regression_model']['scores']
    print(scores)

    assert math.isclose(scores['Explained variance'], 0.531103247696713, abs_tol = 0.00001)
#{'Explained variance': 0.531103247696713, 'Max error': 0.8037485197869163, 'Mean absolute error': 0.2050877131438389, 'MSE': 0.09214453355442812, 'Mean squared log error': 0.005413269099842543, 'R2': 0.5252494593646577, 'eval_time': datetime.datetime(2019, 7, 26, 13, 12, 13, 185343)}



'''A module that writes something somewhere

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

'''
from primrose.base.node import AbstractNode

class AbstractWriter(AbstractNode):
    """write some data somewhere"""
    pass



"""Module with abstract postprocess class to specify interface needed for future postprocesses

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.node import AbstractNode

class AbstractPostprocess(AbstractNode):
    """Postprocess module which must have an postprocess method to send data to an external source"""
    pass



import pytest
from primrose.dag.depth_first_traverser import DepthFirstTraverser
from primrose.dag.config_layer_traverser import ConfigLayerTraverser
from primrose.dag.traverser_factory import TraverserFactory

def test_instantiate():
    assert isinstance(TraverserFactory().instantiate('DepthFirstTraverser', None), DepthFirstTraverser)
    assert isinstance(TraverserFactory().instantiate('ConfigLayerTraverser', None), ConfigLayerTraverser)
    assert isinstance(TraverserFactory().default_traverser(None), ConfigLayerTraverser)



import pytest
from primrose.configuration.configuration import Configuration
from primrose.conditionalpath.simple_switch import SimpleSwitch

def test_destinations_to_prune():
    config = {
        "implementation_config": {
            "reader_config": {
                "conditional_node": {
                    "class": "SimpleSwitch",
                    "path_to_travel": "left",
                    "destinations": ['left', 'right']
                }
            },
            "writer_config": {
                "left": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                },
                "right": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }

    configuration = Configuration(None, is_dict_config=True, dict_config=config)
    node = SimpleSwitch(configuration, "conditional_node")
    to_prune = node.destinations_to_prune()
    assert to_prune == ['right']




import math
import pytest
import pandas as pd
import numpy as np

from primrose.transformers.impute import ColumnSpecificImpute

def test_fit():
    d = {'col1': [1, 2, 3],
         'col2': [4, 6, 11.2],
         'col3': [1, 1, 1],
         'col4': [3, 2, 3],
         'col5': [1, 2, 3],
         'col6': [1, 2, 3],
         'col7': [None, None, None],
         'col8': np.nan}
    df = pd.DataFrame(data=d)

    imputer = ColumnSpecificImpute(columns_to_zero=['col1'],columns_to_mean=['col2'],columns_to_median=['col3'],columns_to_mode=['col4','col7', 'col8'], columns_to_infinity=['col5'],columns_to_neg_infinity=['col6'])

    imputer.fit(df)

    encoder = imputer.encoder
    assert encoder['col1'] == 0
    assert math.isclose(encoder['col2'], 21.2/3, abs_tol=0.001)
    assert encoder['col3'] == 1
    assert encoder['col4'] == 3
    assert math.isclose(encoder['col5'], 999999999, abs_tol=0.001)
    assert math.isclose(encoder['col6'], -999999999, abs_tol=0.001)
    assert encoder['col7'] == 0
    assert encoder['col8'] == 0

def test_fit2():
    d = {'col1': [1, 2, 3],
         'col2': [4, 6, 11.2],
         'col3': [1, 1, 1],
         'col4': [3, 2, 3],
         'col5': [1, 2, 3],
         'col6': [1, 2, 3]}
    df = pd.DataFrame(data=d)

    imputer = ColumnSpecificImpute(columns_to_zero=['col1'],columns_to_mean=['col2', 'col1'],columns_to_median=['col3'],columns_to_mode=['col4'], columns_to_infinity=['col5'],columns_to_neg_infinity=['col6'])

    with pytest.raises(Exception) as e:
        imputer.fit(df)
    assert "There are columns in multiple lists {'col1'}" in str(e)

def test_fit3():
    d = {'col1': [1, 2, 3],
         'col2': [4, 6, 11.2],
         'col3': [1, 1, 1],
         'col4': [3, 2, 3],
         'col5': [1, 2, 3],
         'col6': [1, 2, 3]}
    df = pd.DataFrame(data=d)

    imputer = ColumnSpecificImpute(columns_to_zero=['col1'],columns_to_mean=['JUNK'],columns_to_median=['col3'],columns_to_mode=['col4'], columns_to_infinity=['col5'],columns_to_neg_infinity=['col6'])

    with pytest.raises(Exception) as e:
        imputer.fit(df)
    assert "Unrecognized impute column 'JUNK'" in str(e)

def test_transform():
    df = pd.DataFrame()
    imputer = ColumnSpecificImpute(columns_to_zero=['col1'],columns_to_mean=['col2'],columns_to_median=['col3'],columns_to_mode=['col4'], columns_to_infinity=['col5'],columns_to_neg_infinity=['col6'])
    with pytest.raises(Exception) as e:
        imputer.transform(df)
    assert 'ColumnSpecificImpute must train imputations with fit before calling transform' in str(e)

def test_transform2():
    d = {'col1': [1, np.nan, 3],
         'col2': [4, 6, np.nan]}
    df = pd.DataFrame(data=d)

    imputer = ColumnSpecificImpute(columns_to_zero=['col1'],columns_to_mean=['col2'],columns_to_median=[],columns_to_mode=[], columns_to_infinity=[],columns_to_neg_infinity=[])

    imputer.fit(df)

    transormed_df = imputer.transform(df)

    assert list(transormed_df.col1) == [1, 0, 3]
    assert list(transormed_df.col2) == [4, 6, 5]




"""Module to handle book keeping of data

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os
from collections import defaultdict
import logging
from enum import Enum
import dill

class DataObjectResponseType(Enum):
    """Type of object when getting data from DataObject

    INSTANCE_KEY_VALUE = dictionary of instance_name keys and their data dictionaries: 
        {'instance_name': {'key':value}, 'instance_name2': {'key2':value2}, ... }
        e.g. {'corpus_reader': {'data': dataframe}}
        this is useful if there is a set of upstream data arriving from mulitple sources

    KEY_VALUE = dictonary of data for given instance name: {'key':value}
        e.g. {'data': dataframe} or {'data': dataframe, 'query': 'select * from table'}
        this is useful if there are multiple keys for a given instance_name
        or if you want to explicitly check against expected keys

    VALUE = value only (for 1st or only instance name and for only key): 
        e.g. dataframe
        this is useful if you know a node in DAG has only a single upsteam source and only a single
        value. Readers are often a good example as they typically read in and provide a single data frame

    """
    INSTANCE_KEY_VALUE = "ikv"
    KEY_VALUE = "kv"
    VALUE = "v"

    @staticmethod
    def values():
        ''' list of all the values in the enum'''
        return list(map(lambda t: t.value, DataObjectResponseType))

class DataObject():
    """DataObject: a container for "data" (strings, dicts, arbitrary objects etc)"""

    # when we are storing some basic data, what is key we use?
    DATA_KEY = 'data'

    DEFAULT_RESPONSE_TYPE = DataObjectResponseType.KEY_VALUE.value

    def __init__(self, config):
        """instantiate the DataObject

        Args:
            config (Configuration): Configuration instance

        """
        #assert isinstance(config, Configuration)
        self.config = config
        self.data_dict = defaultdict(dict)

    @staticmethod
    def read_from_cache(filename):
        """restore DatObject from dill-cached file

        Args:
            filename (str): cache filename

        Returns:
            data_object (DataObject): DataObject instance from cache

        """
        assert os.path.exists(filename)
        with open(filename, 'rb') as f:
            data_object = dill.load(f)
            assert isinstance(data_object, DataObject)
            return data_object

    def write_to_cache(self, filename):
        """write data_object (self) to dill-cache

        Returns:
            nothing. Side effect is to cache object to file

        """
        with open(filename, 'wb') as f:
            logging.info("Cache DataObect to " + filename) 
            dill.dump(self, f)

    def __repr__(self):
        """string representation of the class
        
        Returns:
            string representation

        """
        return self.__class__.__name__ + ":" + str(self.data_dict)

    def add(self, requestor, data, key=DATA_KEY, overwrite=False):
        """for requestor's instance_name, set key:data in storage

        Args:
            requestor (Node): is object (model, pipeline, writer etc) that has instance_name attribute
            data (object): some object
            key (string): if not supplied default data key is used

        Returns:
            nothing.
        """
        assert isinstance(key, str)
        if not overwrite and requestor.instance_name in self.data_dict and key in self.data_dict[requestor.instance_name]:
            raise Exception("Key already exists for %s:%s" % (requestor.instance_name, key))

        # as this is  defaultdict(dict) it should update existing dict with new key
        self.data_dict[requestor.instance_name][key] = data

    def get(self, instance_name, pop_data=False, rtype=DEFAULT_RESPONSE_TYPE):
        """get some data from storage, optionally popping it off.

        Args:
            instance_name (str): name of node in DAG
            pop_data (bool): boolean, whether to pop data from storage
            rtype (DataObjectResponseType): DataObjectResponseType value, specifying response type

        Returns:
            data of desired DataObjectResponseType, selected with rtype

        Raises:
            Exception if unrecognixzed rtype or keys

        """
        assert instance_name
        assert isinstance(instance_name, str)

        assert rtype
        if rtype not in DataObjectResponseType.values():
            raise Exception("Unrecognized rtype: %s" % rtype)

        if not instance_name in self.data_dict.keys():
            raise Exception("Key not found: %s" % instance_name)

        d = None
        if pop_data:
            d = self.data_dict.pop(instance_name)
        else:
            d = self.data_dict[instance_name]

        if rtype == DataObjectResponseType.INSTANCE_KEY_VALUE.value:
            return {instance_name: d}
        elif rtype == DataObjectResponseType.KEY_VALUE.value:
            return d
        elif rtype == DataObjectResponseType.VALUE.value:
            # only if there is a dict with single key can we return value
            if isinstance(d, dict):
                keys = list(d.keys())
                if len(keys) == 1:
                    return d[keys[0]]
                else:
                    logging.info("Multiple keys found: %s" % str(keys))
            # otherwise return whatever d is
            return d

    def upstream_keys(self, instance_name, operation_type_filter=None):
        """get list of upstream node names for a given input requestor node

        Args:
            instance_name (str): name of requestor
            operation_type_filter (optional): type of operation type to filter in

        Returns:
            list of keys, if any

        """
        assert instance_name
        keys = self.config.dag.upstream_keys(instance_name)

        if operation_type_filter:
            keys = [k for k in keys if self.config.dag.node_map[k] == operation_type_filter.value]

        return keys

    def get_upstream_data(self, instance_name, pop_data=False, rtype=DEFAULT_RESPONSE_TYPE, operation_type_filter=None):
        """Return data from upstream source(s), choose to pop or not from the dict

        Note:
            returns dictionary, where keys are instance_names and each value is a dictionary.
            However, if
            i) there is only 1 upstream key
            and
            ii) value_only=True
            then return the value only.
            
            This option is useful if you expect 1 upstream source only and it returns
            a single artifact, such as a single dataframe. In that case just the dataframe
            is returned

        Returns:
            object (type depends on DEFAULT_RESPONSE_TYPE)

        Raises:
            Exception if no upstream data found

        """
        assert instance_name
        upstream_keys = self.upstream_keys(instance_name, operation_type_filter=operation_type_filter)

        # While upstream_keys list the upstream sources, it doesn't mean any data were set for them.
        # Thus, we need to see which of these we have data for
        upstream_keys_with_data = set(upstream_keys).intersection(set(self.data_dict.keys()))

        if not upstream_keys_with_data:
            raise Exception("No upstream keys with data found for %s" % instance_name)

        # if there are multiple keys, returning key:value or value makes no sense so ignore rtype
        if len(upstream_keys) > 1:
            return {iname_key: self.get(iname_key, pop_data) for iname_key in upstream_keys}

        # we must have only 1 key now so this is now simple:
        return self.get(upstream_keys[0], pop_data, rtype)

    def get_filtered_upstream_data(self, instance_name, filter_for_key):
        """Upstream data where first level dict keys are first checked for the presence of a filter key
        
        Args:
            instance_name (str): name of instance to look upstream from

            filter_for_key (str): the key data was saved with (not instance name but the data value key)

        Returns:
            dictionary of stored data for that instance if only one matching dict,
            if more than one valid dictionary then return list of dicts,
            None otherwise

        """
        data = self.get_upstream_data(instance_name, pop_data=False,
                                        rtype=DataObjectResponseType.INSTANCE_KEY_VALUE.value)

        data_to_return = []

        for ikey in data:
            if filter_for_key in data[ikey]:
                data_to_return.append(data[ikey])

        if len(data_to_return) == 1:
            return data_to_return[0]

        elif not data_to_return:  # pep8 for empty list checks
            return None

        else:
            return data_to_return



""""Encode all string df columns to numeric labels, inherit test/train split from parent

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""
from primrose.pipelines.train_test_split import TrainTestSplit
from primrose.base.transformer_sequence import TransformerSequence
from primrose.transformers.categoricals import ImplicitCategoricalTransform


class EncodeTrainTestSplit(TrainTestSplit):
    """Encode all string df columns to numeric labels, inherit test/train split from parent"""

    def init_pipeline(self):
        """create the pipeline's TransformerSequence

        Returns:
            a TransformerSequence

        """
        ts = TransformerSequence()
        ts.add(ImplicitCategoricalTransform(self.node_config['target_variable']))
        return ts

    @property
    def first_transformer_in_sequence(self):
        """returns 1st transformer in a sequence

        Returns:
            a transformer

        """
        return self.transformer_sequence.sequence[0]

    def final_data_object_additions(self, data_object):
        """Overload function which adds the label encoder after running fit_transform or transform
        
        Returns:
            data_object (DataObject): instance of DataObject
        
        """
        # add target label encoder to the data object if it's needed downstream
        target_label_encoder = self.first_transformer_in_sequence.target_encoder
        data_object.add(self, target_label_encoder, key='target_encoder')
        return data_object



"""write some data to a local file, serialized with dill

Author(s):
    Mike Skarlinski (michael.skarlinski@weightwatchers.com)

"""
import os
import logging
import dill
import warnings

from primrose.writers.abstract_file_writer import AbstractFileWriter
from primrose.data_object import DataObjectResponseType


class DillWriter(AbstractFileWriter):
    ''' write some specified data object to a dill file '''
    warnings.warn('Use Serializer instead. DillWriter will be deprecated in a future release.', DeprecationWarning)

    def run(self, data_object):
        """serialize some data to a local filesystem using Dill

        Note:
            this will write object under config['key'] in data_object

        Args:
            data_object (DataObject)

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        filename = self.node_config['dir'] + os.path.sep + self.node_config['filename']
        data_key = self.node_config['key']

        logging.info("Saving {} data to {}".format(data_key, filename))

        data_to_write = data_object.get_upstream_data(self.instance_name, pop_data=False, rtype=DataObjectResponseType.KEY_VALUE.value)

        if data_key not in data_to_write:
            raise Exception("Key {} not found inside data_key object".format(data_key))

        with open(filename, 'wb') as f:
            dill.dump(data_to_write[data_key], f)

        terminate = False

        return data_object, terminate




import pytest
import sys
import os
import boto3
from moto import mock_s3
import pandas as pd
from primrose.configuration.configuration import Configuration
from primrose.writers.s3_writer import S3Writer
from primrose.node_factory import NodeFactory
from primrose.data_object import DataObject
from primrose.base.node import AbstractNode

@mock_s3
def test_init_ok():
    config = {
        "implementation_config": {
            "postprocess_config": {
                "nodename": {
                    "class": "TestPostprocess",
                    "key1":"val1",
                    "key2":"val2",
                    "destinations": ["recipe_s3_writer"]
                }
            },
            "writer_config": {
                "recipe_s3_writer": {
                    "class": "S3Writer",
                    "dir": "cache",
                    "key": DataObject.DATA_KEY,
                    "bucket_name": "does_not_exist_bucket_name",
                    "bucket_filename": "does_not_exist.csv"
                }
            }
        }
    }
    class TestPostprocess(AbstractNode):
        @staticmethod
        def necessary_config(node_config):
            return set(['key1', 'key2'])
        def run(self, data_object): return data_object
    NodeFactory().register("TestPostprocess", TestPostprocess)

    #this is to mock out the boto connection
    os.environ["AWS_ACCESS_KEY_ID"] = "fake" 
    os.environ["AWS_SECRET_ACCESS_KEY"] = "fake"
    conn = boto3.resource('s3')
    # We need to create the bucket since this is all in Moto's 'virtual' AWS account
    conn.create_bucket(Bucket='does_not_exist_bucket_name')

    reference_file_path = "test/minimal.csv"

    corpus = pd.read_csv(reference_file_path)

    configuration = Configuration(None, is_dict_config=True, dict_config=config)

    data_object = DataObject(configuration)

    requestor = TestPostprocess(configuration, 'nodename')

    data_object.add(requestor, corpus)

    writer = S3Writer(configuration, 'recipe_s3_writer')
    node_config = {
                    "class": "S3Writer",
                    "dir": "cache",
                    "key": DataObject.DATA_KEY,
                    "bucket_name": "does_not_exist_bucket_name",
                    "bucket_filename": "does_not_exist.csv"
                }
    keys = writer.necessary_config(node_config)
    assert keys is not None
    assert isinstance(keys, set)
    assert len(keys) > 0

    # write to file
    filename = writer._write_locally(data_object)
    assert os.path.exists(filename)

    # check it is same data as expected
    reference = pd.read_csv(reference_file_path)
    just_written = pd.read_csv(filename)

    assert reference.equals(just_written)

    os.remove(filename)

    data_object = writer.run(data_object)
    body = conn.Object('does_not_exist_bucket_name', 'does_not_exist.csv').get()['Body'].read().decode("utf-8")
    assert body == open(reference_file_path).read()


"""write some dataframe to a local CSV file.

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
import os
import logging
from primrose.data_object import DataObjectResponseType
from primrose.writers.abstract_file_writer import AbstractFileWriter

class CsvWriter(AbstractFileWriter):
    """write some dataframe to a local CSV file."""

    def get_optional_config(self):
        '''
        Optionally get kwargs to pass to pandas csv reader.

        Notes:
            kwargs (dict): dictionary of kwargs key-value pairs

        Example:
            "csv_reader": { \
                "class": "CsvReader", \
                "filename": "data/mydata.csv", \
                "kwargs": { \
                "header": None, \
                "sep": ":" \
                } \
            }

        '''
        if 'kwargs' in self.node_config:
            kwargs = self.node_config['kwargs']
            if 'index' not in kwargs:
                kwargs['index'] = False
            return kwargs
        else:
            return {"index": False}

    def run(self, data_object):
        """write some dataframe to a local CSV file

        Returns:
            (tuple): tuple containing:

                data_object (DataObject): instance of DataObject

                terminate (bool): terminate the DAG?

        """
        filename = self.node_config['dir'] + os.path.sep + self.node_config['filename']
        key = self.node_config['key']
        logging.info("Saving %s data to %s", key, filename)
        data_to_write = data_object.get_upstream_data(self.instance_name, pop_data=False, rtype=DataObjectResponseType.KEY_VALUE.value)
        kwargs = self.get_optional_config()
        data_to_write[key].to_csv(filename, **kwargs)
        terminate = False
        return data_object, terminate




import pytest
import sys
import os
import json
from primrose.configuration.configuration import Configuration
from primrose.writers.abstract_file_writer import AbstractFileWriter
from primrose.node_factory import NodeFactory

@pytest.fixture
def mock_env(monkeypatch):
    monkeypatch.setenv("PRIMROSE_EXT_NODE_PACKAGE", "test")

def test_init_error0():
    config = {}
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'Did not find required top-level key implementation_config' in str(e)

def test_init_error1():
    with pytest.raises(Exception) as e:
        Configuration(None)
    if sys.version_info[:2] == (3, 5):
        assert "stat: can't specify None for path argument" in str(e)
    else:
        assert 'stat: path should be string, bytes, os.PathLike or integer, not NoneType' in str(e)

def test_init_error2():
    with pytest.raises(Exception) as e:
        Configuration(config_location="junky_path_does_not_exist")
    assert 'config file at: junky_path_does_not_exist not found' in str(e)

def test_init_error3():
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=None)
    assert 'expected dict_config was None' in str(e)

def test_init_error4():
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config="1234")
    assert 'did not receive expected dict_config' in str(e)

def test_init_error5():
    config = {  
        "implementation_config": {
            "k1": {"a":1},
            "k2": {"a":2}
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'Operations must all have unique names in the configuration' in str(e)

def test_init_error6():
    with pytest.raises(Exception) as e:
        filename = "test/test_dupe_toplevel.json"
        Configuration(config_location=filename, is_dict_config=False, dict_config=None)
    assert "duplicate key: 'model_config'" in str(e)

def test_init_error7():
    config = {  
        "junk": {}
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'Unsupported top-level key: junk. Supported keys are [\'metadata\', \'implementation_config\']' in str(e)

def test_init_error8():
    with pytest.raises(ValueError) as e:
        Configuration('test/tennis.csv')
    assert 'config file at: test/tennis.csv has improper extension type - please use a .json or .yml file' in str(e)

def test_metadata():
    config = {
        "metadata":{"k":"v"},
        "implementation_config":{}
    }
    c = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert c.config_metadata["k"] == "v"

def test_init_ok():
    config = {"implementation_config":{}}
    c = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert c.config_string
    assert c.config_hash

def test_init_ok2():
    config = {
        "metadata": {
            "traverser": "DepthFirstTraverser"
        },
        "implementation_config": {
            "reader_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "some/path/to/file",
                    "destinations": []
                }
            }
        }
    }
    c = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert c.config_string
    assert c.config_hash == 'b434870806fe0c61ddf2b417ae62c1be519b069fb6e12572f4ff8143f98086ff'

def test_unrecognized():
    config = {
        "implementation_config": {
            "junky_config": {
                "csv_reader": {
                    "class": "CsvReader",
                    "filename": "some/path/to/file",
                    "destinations": []
                }
            }
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'Unspported operation: junky_config' in str(e)

def test_nopipelinekey():
    config = {
        "implementation_config":{
            "pipeline_config": {
                "corpus_pipeline": {
                    "destinations": ["recipe_name_model"]
                }
            }
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'No class key found in pipeline_config.corpus_pipeline' in str(e)

def test_metadata2():
    config = {
        "metadata": {
            "traverser": "doesnotexist"
        },
        "implementation_config":{
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'doesnotexist is not a valid and/or registered Traverser' in str(e)

def test_check_metadata():
    config = {
        "metadata": {
            "data_object": {
                "read_from_cache": True
            }
        },
        "implementation_config":{
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "metadata.data_object: if read_from_cache==true, you must set 'read_filename'" in str(e)

def test_check_metadata2():
    config = {
        "metadata": {
            "data_object": {
                "read_from_cache": True,
                "read_filename": "path/does/not/exist"
            }
        },
        "implementation_config":{
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "Invalid metadata.data_object.read_filename: path/does/not/exist" in str(e)

def test_check_metadata3():
    config = {
        "metadata": {
            "data_object": {
                "write_to_cache": True
            }
        },
        "implementation_config":{
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "metadata.data_object: if write_to_cache==true, you must set 'write_filename'" in str(e)

def test_check_setions():
    config = {
        "metadata": {
            "section_registry": [
                "phase_1"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": []
                }
            }
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "Following sections from metadata were not found implementation: {'phase_1'}" in str(e)

def test_check_setions2():
    config = {
        "metadata": {
            "section_registry": [
                "reader_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": []
                }
            },
            "writer_config": {}
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "Following sections from implementation were not found in metadata: {'writer_config'}" in str(e)

def test_check_setions3():
    config = {
        "metadata": {
            "section_registry": [
                "reader_config",
                "writer_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": []
                }
            },
            "writer_config": {}
        }
    }
    # this should not raise an exception
    Configuration(config_location=None, is_dict_config=True, dict_config=config)

def test_check_setions4():
    config = {
        "metadata": {
            "section_registry": [
                "phase1",
                "phase2"
            ]
        },
        "implementation_config": {
            "phase1": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": []
                }
            },
            "phase2": {}
        }
    }
    # this should not raise an exception
    Configuration(config_location=None, is_dict_config=True, dict_config=config)

def test_sections_in_order1():
    config = {
        "metadata": {
            "section_registry": [
                "phase1",
                "phase2"
            ]
        },
        "implementation_config": {
            "phase1": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": []
                }
            },
            "phase2": {}
        }
    }
    config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    sections, source = config.sections_in_order()
    assert sections == ["phase1", "phase2"]
    assert source == "section_registry"

def test_sections_in_order2():
    config = {
        "metadata": {
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": []
                }
            },
            "writer_config": {

            }
        }
    }
    config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    sections, source = config.sections_in_order()
    assert sections == ["reader_config", "writer_config"]
    assert source == 'default'

def test_sections_in_order3():
    config = {
        "metadata": {
            "section_run": [
                "writer_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/tennis.csv",
                    "destinations": ["recipe_csv_writer"]
                }
            },
            "writer_config": {
                "recipe_csv_writer": {
                    "class": "CsvWriter",
                    "key":"test_data",
                    "dir": "cache",
                    "filename": "unittest_similar_recipes.csv"
                }
            }
        }
    }
    config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    sections, source = config.sections_in_order()
    assert sections == ["writer_config"]
    assert source == "section_run"

def test_registration():
    class TestWriterUnreg(AbstractFileWriter):
        def __init__(self, configuration, instance_name):
            pass
        def write(self, data_dict):
            pass
    
    config = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestWriterUnreg",
                    "destinations": []
                }
            }
        }
    }
    with pytest.raises(Exception) as e:
        Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert 'Cannot register node class TestWriterUnreg' in str(e)


def test_comments_in_json():
    # should not raise exception even though there are comments in the JSON
    config = Configuration(config_location="test/config_with_comments.json")
    assert list(config.config.keys()) == ['reader_config', 'writer_config']
    assert config.config['reader_config']['read_data']['class'] == 'CsvReader'
    assert list(config.config['reader_config']['read_data']['destinations']) == ['write_output']
    #"{'reader_config': {'read_data': {'class': 'CsvReader', 'filename': 'data/tennis.csv', 'destinations': ['write_output']}}, 'writer_config': {'write_output': {'class': 'CsvWriter', 'key': 'data', 'dir': 'cache', 'filename': 'tennis_output.csv'}}}"


def test_perform_any_config_fragment_substitution_bad():
    config_str = """
    {
        {% include "does/not/exist" %}
        "implementation_config": {
        }
    }
    """
    with pytest.raises(Exception) as e:
        Configuration.perform_any_config_fragment_substitution(config_str)
    assert 'Substitution files do not exist: does/not/exist' in str(e)


def test_perform_any_config_fragment_substitution():
    config_str = """
    {
        {% include "test/metadata_fragment.json" %}
        "implementation_config": {
            {% include "test/read_write_fragment.json" %}
        }
    }
    """
    final_str = Configuration.perform_any_config_fragment_substitution(config_str)
    expected = """
    {
        "metadata":{},
        "implementation_config": {

        "reader_config": {
            "read_data": {
                "class": "CsvReader",
                "filename": "data/tennis.csv",
                "destinations": [
                    "write_output"
                ]
            }
        },
        "writer_config": {
            "write_output": {
                "class": "CsvWriter",
                "key": "data",
                "dir": "cache",
                "filename": "tennis_output.csv"
            }
        }
        
        }
    }
    """
    assert json.loads(final_str) == json.loads(expected)

def test_yaml_config1():
    config_yaml = Configuration(config_location='test/hello_world_tennis.yml')
    config_json = Configuration(config_location='test/hello_world_tennis.json')
    assert config_yaml.config_hash == config_json.config_hash

def test_yaml_config2():
    c = Configuration('test/config_substitution.yml')
    assert c.config_string
    assert c.config_hash

def test_yaml_perform_any_config_fragment_substitution():
    config_str = """
{% include "test/metadata_fragment.yml" %}
implementation_config:
{% include "test/read_write_fragment.yml" %}
    """
    final_str = Configuration.perform_any_config_fragment_substitution(config_str)
    expected = """
metadata: {}
implementation_config:
  reader_config:
    read_data:
      class: CsvReader
      destinations:
      - write_output
      filename: data/tennis.csv
  writer_config:
    write_output:
      class: CsvWriter
      dir: cache
      filename: tennis_output.csv
      key: data
    """
    assert final_str == expected

def test_class_prefix(mock_env):
    config_path = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "class_prefix": "test/ext_node_example.py",
                    "destinations": []
                }
            }
        }
    }
    config_dot = {
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "class_prefix": "test.ext_node_example",
                    "destinations": []
                }
            }
        }
    }
    for config in [config_path, config_dot]:
        config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
        assert config.config_string
        assert config.config_hash
        NodeFactory().unregister('TestExtNode')
        

def test_class_package(mock_env):
    config_path = {
        "metadata": {
            "class_package": "test"
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "destinations": []
                }
            }
        }
    }
    config_full_path = {
        "metadata": {
            "class_package": "test/ext_node_example.py"
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "destinations": []
                }
            }
        }
    }
    config_full_dot = {
        "metadata": {
            "class_package": "test"
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "class_prefix": "ext_node_example",
                    "destinations": []
                }
            }
        }
    }
    for config in [config_full_path, config_path, config_full_dot]:
        config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
        assert config.config_string
        assert config.config_hash
        NodeFactory().unregister('TestExtNode')


def test_env_override_class_package(mock_env):
    config = {
        "metadata": {
            "class_package": "junk"
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "destinations": []
                }
            }
        }
    }
    config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert config.config_string
    assert config.config_hash
    NodeFactory().unregister('TestExtNode')


def test_incorrect_class_package():
    config = {
        "metadata": {
            "class_package": "junk"
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "destinations": []
                }
            }
        }
    }
    with pytest.raises(Exception) as e:
        config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "Cannot register node class TestExtNode" in str(e)

def test_incorrect_class_package2():
    config = {
        "metadata": {
            "class_package": "test"
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "TestExtNode",
                    "class_prefix": "junk",
                    "destinations": []
                }
            }
        }
    }
    with pytest.raises(Exception) as e:
        config = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    assert "Cannot register node class TestExtNode" in str(e)




import pytest
from primrose.configuration.configuration import Configuration
from primrose.dag_runner import DagRunner
import os

def test_run():
    config = {
        "metadata": {
            "section_registry": [
                "reader_config",
                "pipeline_config",
                "model_config",
                "dataviz_config"
            ]
        },
        "implementation_config": {
            "reader_config": {
                "read_data": {
                    "class": "CsvReader",
                    "filename": "test/unclustered.csv",
                    "destinations": [
                        "normalize_data"
                    ]
                }
            },
            "pipeline_config": {
                "normalize_data": {
                    "class": "SklearnPreprocessingPipeline",
                    "operations": [
                        {"class":"preprocessing.StandardScaler", "columns": ["x", "y"], "args": {"with_mean": True, "with_std": True}},
                    ],
                    "is_training": True,
                    "training_fraction": 0.65,
                    "seed": 42,
                    "destinations": [
                        "cluster_model"
                    ]
                }
            },
            "model_config": {
                "cluster_model":{
                    "class": "SklearnClusterModel",
                    "mode": "train",
                    "features": ["x","y"],
                    "model": {"class": "cluster.KMeans", "args": {"n_clusters": 6, "random_state": 42}},
                    "destinations": [
                        "cluster_plotter"
                    ]
                }
            },
            "dataviz_config": {
                "cluster_plotter": {
                    "class": "ClusterPlotter",
                    "id_col": "predictions",
                    "filename": "clusters.png",
                    "title": "Results of KMeans(k=6)",
                    "destinations": []
                }
            }
        }
    }
    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)

    fname = "clusters.png"

    if os.path.exists(fname):
        os.remove(fname)

    DagRunner(configuration).run()

    assert os.path.exists(fname)

    if os.path.exists(fname):
        os.remove(fname)



"""A primrose model based around a sklearn model

Author(s):
    Carl Anderson (carl.anderson@weightwatchers.com)

"""
from primrose.base.model import AbstractModel
import importlib
import logging
from sklearn.metrics import roc_curve, auc
from sklearn import metrics
import datetime

class SklearnModel(AbstractModel):

    def __init__(self, configuration, instance_name):
        """A Sklearn-based model to train, evaluate and predict on dataframe feature data

        Args:
            configuration (Configuration): user defined configuration object
            instance_name (str): name to identify the correct configuration key

        """
        super(SklearnModel, self).__init__(configuration, instance_name)
        self.model = None
        self.scores = {}

    def load_model(self, data_object):
        """finds an upstream sklearn mode

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            Sklearn model

        """

        # check for a valid pipeline object inside the data object
        model = data_object.get_filtered_upstream_data(self.instance_name, filter_for_key='reader_data')['reader_data']

        if model is None:
            raise Exception('No valid upstream model found in the data object for model.')

        logging.info('Model successfully loaded from upstream data_object.')

        return model

    @staticmethod
    def _instantiate_model(classname, args):
        """instantiates a Sklearn model from class name

        Args:
            classname (str): name of class without initial `sklearn.`, e.g. `cluster.KMeans`

            args (dict): dictionary of arguments/parameters for model instantiation

        Returns:
            instance of a sklearn mode

        """
        sk_module_name, sk_model_name = classname.split('.')
        sk_module = importlib.import_module('sklearn.{}'.format(sk_module_name))

        try:
            m = getattr(sk_module, sk_model_name)
        except AttributeError:
            raise Exception('Sklearn model {} not found in {} module'.format(sk_model_name, sk_module_name))
        logging.info("Instantiating model " + classname)
        if args:
            return m(**args)
        else:
            return m()

    @staticmethod
    def _get_roc_score(actual_values, model_probability):
        """Get (binary) ROC store using actual target variables and predicted probabilities

            Assumes that the model is predicting 1 as the minority class

        Args:
            actual_values: series with correct target variable assignments
            model_probability: probability of target variable assignment from the model

        Returns: AUC for a binary classifier

        """
        fpr, tpr, _ = roc_curve(actual_values, [y[1] for y in model_probability])
        return auc(fpr, tpr)

    def _get_data(self, data_object):
        """get the upstream data, split into X and Y from config, return data frame

        Returns:
            dataframe (dataframe)

        """

        # get upstream data dict which contains the subkey data_test or data_train
        data = (data_object.get_filtered_upstream_data(self.instance_name, 'data_test') or 
                data_object.get_filtered_upstream_data(self.instance_name, 'data_train'))

        X_train = None
        if 'data_train' in data:
            X_train = data['data_train']
            if 'features' in self.node_config:
                X_train = X_train[self.node_config['features']]

        y_train = None
        if 'target_train' in data:
            y_train = data['target_train']

        X_test = None
        if 'data_test' in data:
            X_test = data['data_test']
            if 'features' in self.node_config:
                X_test = X_test[self.node_config['features']]

        y_test = None
        if 'target_test' in data:
            y_test = data['target_test']

        def size(obj):
            if obj is None:
                return None
            return(str(obj.shape))

        logging.info("Get_data, X_train %s, y_train %s, X_test %s, y_test %s", size(X_train), size(y_train), size(X_test), size(y_test))

        return X_train, y_train, X_test, y_test

    def train_model(self, data_object):
        """train the model

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            data_object (DataObject): instance of DataObject

        """
        self.X_train, self.y_train, self.X_test, self.y_test = self._get_data(data_object)

        args = None
        if 'args' in self.node_config['model']:
            args = self.node_config['model']['args']

        self.model = self._instantiate_model(self.node_config['model']['class'], args)

        logging.info("Fitting model")
        self.fit_training_data() #delegate down as args to self.model.fit() vary by model

        data_object.add(self, self.model, 'model')

        return data_object

    def _make_predictions(self, data_object):
        """Make predictions from X_test

        Args:
            data_object (DataObject): instance of DataObject

        Returns:
            nothing. Predictions stored in self.predictions

        """
        if hasattr(self, 'predictions') and self.predictions is not None:
            return

        if self.model is None:
            self.model = self.load_model(data_object)

        self.X_train, self.y_train, self.X_test, self.y_test = self._get_data(data_object)

        logging.info("Making predictions with model")
        self.predictions = self.model.predict(self.X_test)

    def predict(self, data_object, load_model=False, use_serial=False):
        """Predict y_test from X_test

        Args:
            data_object (DataObject): instance of DataObject

            load_model: load model object from gcs or not

        Returns: 
            data_object (DataObject): instance of DataObject

        """
        self._make_predictions(data_object)

        data = self.X_test
        data['predictions'] = self.predictions
        if self.y_test is not None:
            data['actual'] = self.y_test

        data_object.add(self, data)
        return data_object

    def eval_model(self, data_object, load_model=False):
        """evalute a model by getting the scores

        Returns:
            data_object (DataObject): instance of DataObject

        """
        self._make_predictions(data_object)

        logging.info("Evaluating metrics")
        scores = self.get_scores() #delegate down as appropriate metrics vary by model
        scores['eval_time'] = datetime.datetime.now()
        logging.info("Scores" + str(scores))

        data_object.add(self, scores, "scores")
        return data_object


    @staticmethod
    def evaluate_no_ground_truth_classifier_metrics(X, labels):
        """Compute a set of metric for a classifier where there is no ground truth

        Args:
            X (datafframe): the data
            labels: the predicted classes

        Returns:
            dictionary of score name: value

        """
        scores = {}
        scores['Silhouette Score']         = metrics.silhouette_score(X, labels, metric='euclidean')
        scores['Calinski Harabasz Score']  = metrics.calinski_harabasz_score(X, labels)
        scores['Davies Bouldin Score']     = metrics.davies_bouldin_score(X, labels)
        return scores

    @staticmethod
    def evaluate_regression_metrics(actual, predictions):
        """compute set of metrics for a regression

        Args:
            actual: vector of actual values
            predictions: vector of predictions

        Returns:
            dictionary of score name: value

        """
        scores = {}
        scores['Explained variance']       = metrics.explained_variance_score(actual, predictions)
        scores['Max error']                = metrics.max_error(actual, predictions)
        scores['Mean absolute error']      = metrics.mean_absolute_error(actual, predictions)
        scores['MSE']                      = metrics.mean_squared_error(actual, predictions)
        scores['Mean squared log error']   = metrics.mean_squared_log_error(actual, predictions)
        scores['Mean absolute error']      = metrics.median_absolute_error(actual, predictions)
        scores['R2']                       = metrics.r2_score(actual, predictions)
        return scores


import pytest
import os
import sys
import sqlite3
from primrose.configuration.configuration import Configuration
from primrose.readers.sqlite_reader import SQLiteReader
from primrose.data_object import DataObject, DataObjectResponseType

def test_necessary_config():
    assert isinstance(SQLiteReader.necessary_config({}), set)
    assert len(SQLiteReader.necessary_config({})) == 2

def test_read():
    config = {
        "implementation_config":{
            "reader_config": {
                "mynode": {
                    "class": "SQLiteReader",
                    "filename": "test/test_sqlite.db",
                    "query_json": [
                        {"query": "test/test_sqlite.sql"}
                    ],
                    "destinations": []
                }
            }
        }
    }

    filename = "test/test_sqlite.db"
    if os.path.exists(filename):
        os.remove(filename)

    conn = sqlite3.connect(filename)
    c = conn.cursor()
    c.execute("create table test(firstname text, lastname text);")
    c.execute("insert into test(firstname, lastname) values('joe', 'doe'), ('mary','poppins');")
    conn.commit()
    conn.close()

    configuration = Configuration(config_location=None, is_dict_config=True, dict_config=config)
    reader = SQLiteReader(configuration, "mynode")

    data_object = DataObject(configuration)

    data_object, terminate = reader.run(data_object)
    assert not terminate
    dd = data_object.get('mynode', rtype=DataObjectResponseType.KEY_VALUE.value)
    assert "query_0" in dd
    df = dd['query_0']
    assert df is not None
    assert df.shape == (2, 2)

    if os.path.exists(filename):
        os.remove(filename)



from setuptools import setup, find_packages
import os

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

local_data_files = [os.path.join('data', f) for f in os.listdir('data')]
local_config_files = [os.path.join('config', f) for f in os.listdir('config')]

setup(name='primrose',
      version='1.0.11',
      description='Primrose: a framework for simple, quick modeling deployments',
      url='https://github.com/ww-tech/primrose',
      author='Carl Anderson',
      author_email='carl.anderson@weightwatchers.com',
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='Apache 2.0',
      zip_safe=False,
      install_requires=required,
      packages=find_packages(),
      data_files=[('data', local_data_files),
                  ('config', local_config_files)],
      include_package_data=True,
      classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent"
      ],
      project_urls={
        'Documentation': 'https://ww-tech.github.io/primrose/',
        'Source': 'https://github.com/ww-tech/primrose',
      },
      entry_points={"console_scripts": ["primrose = primrose.__init__:cli"]},
      extras_require={
        'postgres': ["psycopg2>=2.8.3", "psycopg2_binary>=2.8.2"],
        'plotting': ["pygraphviz>=1.5"],
        'R': ["rpy2>=2.9.1,<=3.2.7"],
        'all': ["psycopg2>=2.8.3", "psycopg2_binary>=2.8.2", "pygraphviz>=1.5", "rpy2>=2.9.1,<=3.2.7"]
      }
    )


"""Module with abstract transformer class to specify interface needed for transformers

Author(s):
    Michael Skarlinski (michael.skarlinski@weightwatchers.com)

    Carl Anderson (carl.anderson@weightwatchers.com)

"""

from abc import ABC, abstractmethod

class AbstractTransformer(ABC):
    """Serializable object that can be string together within a pipeline"""

    @abstractmethod
    def fit(self, data):
        """User implements fit operation on a single data element from a data_object
        
        Args:
            data (object): some data

        Returns:
            data

        """
        pass # pragma: no cover

    @abstractmethod
    def transform(self, data):
        """User implements internal transform function which operates on a single data element from a data_object
        
        Args:
            data (object): input data

        Returns:
            data, transformed

        """
        return data # pragma: no cover

    def fit_transform(self, data):
        """fit then transform data

        Args:
            data (object): input data

        Returns:
            data, transformed

        """
        self.fit(data)
        return self.transform(data)